"""
ImproperAngleConstraints contains classes for all constraint's related to
improper angles between bonded atoms.

.. inheritance-diagram:: fullrmc.Constraints.ImproperAngleConstraints
    :parts: 1
"""
# standard libraries imports
from __future__ import print_function

# external libraries imports
import numpy as np

# fullrmc imports
from ..Globals import INT_TYPE, FLOAT_TYPE, PI, LOGGER
from ..Globals import str, long, unicode, bytes, basestring, range, xrange, maxint
from ..Core.Collection import is_number, is_integer, raise_if_collected, reset_if_collected_out_of_date
from ..Core.Collection import get_caller_frames
from ..Core.Constraint import Constraint, SingularConstraint, RigidConstraint
from ..Core.dihedral_angles import full_dihedral_angles_coords




class DihedralAngleConstraint(RigidConstraint, SingularConstraint):
    """
    Dihedral angle is defined between two intersecting planes formed
    with defined atoms. Dihedral angle constraint can control up to three
    angle shells at the same times.


    +---------------------------------------------------------------------------+
    |.. figure:: dihedralSketch.png                                             |
    |   :width: 312px                                                           |
    |   :height: 200px                                                          |
    |   :align: center                                                          |
    |                                                                           |
    |   Dihedral angle sketch defined between two planes formed with four atoms.|
    +---------------------------------------------------------------------------+


     .. raw:: html

        <iframe width="560" height="315"
        src="https://www.youtube.com/embed/1wUSYcNygd4"
        frameborder="0" allowfullscreen>
        </iframe>


    :Parameters:
        #. rejectProbability (Number): rejecting probability of all steps
           where standardError increases. It must be between 0 and 1 where 1
           means rejecting all steps where standardError increases and 0
           means accepting all steps regardless whether standardError increases or not.

    .. code-block:: python

        ## Butane (BUT) molecule sketch
        ##
        ##       H13  H22  H32  H43
        ##        |    |    |    |
        ## H11---C1---C2---C3---C4---H41
        ##        |    |    |    |
        ##       H12  H21  H31  H42
        ##

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Constraints.DihedralAngleConstraints import DihedralAngleConstraint

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # create and add constraint
        DAC = DihedralAngleConstraint()
        ENGINE.add_constraints(DAC)

        # define intra-molecular dihedral angles
        DAC.create_angles_by_definition( anglesDefinition={"BUT": [ ('C1','C2','C3','C4', 40,80, 100,140, 290,330), ] })


    """

    def __init__(self, rejectProbability=1):
        # initialize constraint
        RigidConstraint.__init__(self, rejectProbability=rejectProbability)
        # set atomsCollector data keys
        self._atomsCollector.set_data_keys( ['dihedralMap','otherMap'] )
        # init angles data
        self.__anglesList = [[],[],[],[],[],[],[],[],[],[]]
        self.__angles     = {}
        # set computation cost
        self.set_computation_cost(3.0)
        # create dump flag
        self.__dumpAngles = True
        # set frame data
        FRAME_DATA = [d for d in self.FRAME_DATA]
        FRAME_DATA.extend(['_DihedralAngleConstraint__anglesList',
                           '_DihedralAngleConstraint__angles',] )
        RUNTIME_DATA = [d for d in self.RUNTIME_DATA]
        RUNTIME_DATA.extend( [] )
        object.__setattr__(self, 'FRAME_DATA',   tuple(FRAME_DATA) )
        object.__setattr__(self, 'RUNTIME_DATA', tuple(RUNTIME_DATA) )

    @property
    def anglesList(self):
        """ Improper angles list."""
        return self.__anglesList

    @property
    def angles(self):
        """ Angles dictionary for every and each atom."""
        return self.__angles

    def _on_collector_reset(self):
        self._atomsCollector._randomData = set([])

    def listen(self, message, argument=None):
        """
        Listens to any message sent from the Broadcaster.

        :Parameters:
            #. message (object): Any python object to send to constraint's
               listen method.
            #. argument (object): Any type of argument to pass to the listeners.
        """
        if message in("engine set","update boundary conditions",):
            self.set_angles(anglesList=self.__anglesList, tform=False)
            # reset constraint is called in set_angles

    #@raise_if_collected
    def set_angles(self, anglesList, tform=True):
        """
        Sets the angles dictionary by parsing the anglesList list.
        All angles are in degrees. Dihedral angle can control up to three
        angle shells at the same times defined using three different lower
        and upper angle bounds simulating three different dihedral
        potential energy minimums. Dihedral angles are defined from 0 to
        360 degrees. Shell's lower and  upper bound defines a dihedral angle
        clockwise. Therefore in order to take into considerations the limits
        at 0 and 360 degrees, lower bound is allowed to be higher than the
        higher bound.\n
        e.g. (50, 100) is a dihedral shell defined in the angle range between 50 and
        100 degrees. But (100, 50) dihedral shell is defined between 100 to 360 degrees
        and wraps the range from 0 to 100. (50, 100) and (100, 50) are complementary
        and cover the whole range from 0 to 360 deg. \n

        :Parameters:
            #. anglesList (list): The angles list definition.

               tuples format: every item must be a list of ten items.\n
               #. First atom index of the first plane.
               #. Second atom index of the first plane and
                  first atom index of the second plane.
               #. Third atom index of the first plane and second
                  atom index of the second plane.
               #. Fourth atom index of the second plane.
               #. Minimum lower limit of the first shell or
                  minimum angle allowed in degrees which later will be
                  converted to rad.
               #. Maximum upper limit of the first shell or
                  maximum angle allowed in degrees which later will be
                  converted to rad.
               #. Minimum lower limit of the second shell or
                  minimum angle allowed in degrees which later will be
                  converted to rad.
               #. Maximum upper limit of the second shell or
                  maximum angle allowed in degrees which later will be
                  converted to rad.
               #. Minimum lower limit of the third shell or
                  minimum angle allowed in degrees which later will be
                  converted to rad.
               #. Maximum upper limit of the third shell or
                  maximum angle allowed in degrees.\n\n

               ten vectors format: every item must be a list of five items.\n
               #. List containing first atoms index of the first plane.
               #. List containing second atoms index of the
                  first plane and first atoms index of the second plane.
               #. List containing third atoms indexes of the
                  first plane and second atoms index of the second plane.
               #. List containing fourth atoms index of the second plane.
               #. List containing minimum lower limit of the first shell
                  or minimum angle allowed in degrees which later will be
                  converted to rad.
               #. List containing maximum upper limit of the first shell
                  or maximum angle allowed in degrees which later will be
                  converted to rad.
               #. List containing minimum lower limit of the second
                  shell or minimum angle allowed in degrees which later will be
                  converted to rad.
               #. List containing maximum upper limit of the second
                  shell or maximum angle allowed in degrees which later will be
                  converted to rad.
               #. List containing minimum lower limit of the third
                  shell or minimum angle allowed in degrees which later will be
                  converted to rad.
               #. List containing maximum upper limit of the third shell
                  or maximum angle allowed in degrees which later will be
                  converted to rad.

           #. tform (boolean): set whether given anglesList follows tuples format, If not
              then it must follow the ten vectors one.

        **N.B.** Defining three shells boundaries is mandatory. In case fewer than three
        shells is needed, it suffices to repeat one of the shells boundaries.\n
        e.g. ('C1','C2','C3','C4', 40,80, 100,140, 40,80), in the herein definition the
        last shell is a repetition of the first which means only two shells are defined.
        """
        # check if bondsList is given
        if anglesList is None:
            anglesList = [[],[],[],[],[],[],[],[],[],[]]
            tform      = False
        elif len(anglesList) == 10 and len(anglesList[0]) == 0:
            tform     = False
        if self.engine is None:
            self.__anglesList = anglesList
            self.__angles     = {}
        else:
            NUMBER_OF_ATOMS   = self.engine.get_original_data("numberOfAtoms")
            oldAnglesList = self.__anglesList
            oldAngles     = self.__angles
            self.__anglesList = [np.array([], dtype=INT_TYPE),
                                 np.array([], dtype=INT_TYPE),
                                 np.array([], dtype=INT_TYPE),
                                 np.array([], dtype=INT_TYPE),
                                 np.array([], dtype=FLOAT_TYPE),
                                 np.array([], dtype=FLOAT_TYPE),
                                 np.array([], dtype=FLOAT_TYPE),
                                 np.array([], dtype=FLOAT_TYPE),
                                 np.array([], dtype=FLOAT_TYPE),
                                 np.array([], dtype=FLOAT_TYPE)]
            self.__angles     = {}
            self.__dumpAngles = False
            # build angles
            try:
                if tform:
                    for angle in anglesList:
                        self.add_angle(angle)
                else:
                    for idx in xrange(len(anglesList[0])):
                        angle = [anglesList[0][idx], anglesList[1][idx],
                                 anglesList[2][idx], anglesList[3][idx],
                                 anglesList[4][idx], anglesList[5][idx],
                                 anglesList[6][idx], anglesList[7][idx],
                                 anglesList[8][idx], anglesList[9][idx]]
                        self.add_angle(angle)
            except Exception as e:
                self.__dumpAngles = True
                self.__anglesList = oldAnglesList
                self.__angles     = oldAngles
                LOGGER.error(e)
                import traceback
                raise Exception(traceback.format_exc())
            self.__dumpAngles = True
            # finalize angles
            for idx in xrange(NUMBER_OF_ATOMS):
                self.__angles[INT_TYPE(idx)] = self.__angles.get(INT_TYPE(idx), {"idx2":[],"idx3":[],"idx4":[],"dihedralMap":[],"otherMap":[]}  )
        # dump to repository
        self._dump_to_repository({'_DihedralAngleConstraint__anglesList' :self.__anglesList,
                                  '_DihedralAngleConstraint__angles'     :self.__angles})
        # reset constraint
        self.reset_constraint()

    #@raise_if_collected
    def add_angle(self, angle):
        """
        Add a single angle to the list of constraint angles. All angles are in degrees.

        :Parameters:
            #. angle (list): The angle list of ten items.\n
               #. First atom index of the first plane.
               #. Second atom index of the first plane and first
                  atom index of the second plane.
               #. Third atom index of the first plane and second
                  atom index of the second plane.
               #. Fourth atom index of the second plane.
               #. Minimum lower limit of the first shell or minimum
                  angle allowed in degrees which later will be
                  converted to rad.
               #. Maximum upper limit of the first shell or maximum
                  angle allowed in degrees which later will be
                  converted to rad.
               #. Minimum lower limit of the second shell or minimum
                  angle allowed in degrees which later will be
                  converted to rad.
               #. Maximum upper limit of the second shell or maximum
                  angle allowed in degrees which later will be
                  converted to rad.
               #. Minimum lower limit of the third shell or minimum
                  angle allowed in degrees which later will be
                  converted to rad.
               #. Maximum upper limit of the third shell or maximum
                  angle allowed in degrees which later will be
                  converted to rad.
        """
        assert self.engine is not None, LOGGER.error("setting an angle is not allowed unless engine is defined.")
        NUMBER_OF_ATOMS   = self.engine.get_original_data("numberOfAtoms")
        assert isinstance(angle, (list, set, tuple)), LOGGER.error("anglesList items must be lists")
        assert len(angle)==10, LOGGER.error("anglesList items must be lists of 10 items each")
        idx1, idx2, idx3, idx4, lower1, upper1, lower2, upper2, lower3, upper3 = angle
        assert is_integer(idx1), LOGGER.error("angle first item must be an integer")
        idx1 = INT_TYPE(idx1)
        assert is_integer(idx2), LOGGER.error("angle second item must be an integer")
        idx2 = INT_TYPE(idx2)
        assert is_integer(idx3), LOGGER.error("angle third item must be an integer")
        idx3 = INT_TYPE(idx3)
        assert is_integer(idx4), LOGGER.error("angle fourth item must be an integer")
        idx4 = INT_TYPE(idx4)
        assert idx1>=0, LOGGER.error("angle first item must be positive")
        assert idx1<NUMBER_OF_ATOMS, LOGGER.error("angle first item atom index must be smaller than maximum number of atoms")
        assert idx2>=0, LOGGER.error("angle second item must be positive")
        assert idx2<NUMBER_OF_ATOMS, LOGGER.error("angle second item atom index must be smaller than maximum number of atoms")
        assert idx3>=0, LOGGER.error("angle third item must be positive")
        assert idx3<NUMBER_OF_ATOMS, LOGGER.error("angle third item atom index must be smaller than maximum number of atoms")
        assert idx4>=0, LOGGER.error("angle fourth item must be positive")
        assert idx4<NUMBER_OF_ATOMS, LOGGER.error("angle atom index must be smaller than maximum number of atoms")
        assert idx1!=idx2, LOGGER.error("angle second items can't be the same")
        assert idx1!=idx3, LOGGER.error("angle third items can't be the same")
        assert idx1!=idx4, LOGGER.error("angle fourth items can't be the same")
        assert idx2!=idx3, LOGGER.error("angle second and third items can't be the same")
        assert idx2!=idx4, LOGGER.error("angle second and fourth items can't be the same")
        assert idx3!=idx4, LOGGER.error("angle third and fourth items can't be the same")
        assert is_number(lower1), LOGGER.error("angle fifth item must be a number")
        lower1 = FLOAT_TYPE(lower1)
        assert is_number(upper1), LOGGER.error("angle sixth item must be a number")
        upper1 = FLOAT_TYPE(upper1)
        assert lower1>=0, LOGGER.error("angle fifth item must be bigger or equal to 0 deg.")
        assert lower1<=360, LOGGER.error("angle fifth item must be smaller or equal to 360 deg.")
        assert upper1>=0, LOGGER.error("angle sixth item must be bigger or equal to 0 deg.")
        assert upper1<=360, LOGGER.error("angle sixth item must be smaller or equal to 360 deg.")
        #lower1 *= FLOAT_TYPE( PI/FLOAT_TYPE(180.) )
        #upper1 *= FLOAT_TYPE( PI/FLOAT_TYPE(180.) )
        assert is_number(lower2), LOGGER.error("angle seventh item must be a number")
        lower2 = FLOAT_TYPE(lower2)
        assert is_number(upper2), LOGGER.error("angle eights item must be a number")
        upper2 = FLOAT_TYPE(upper2)
        assert lower2>=0, LOGGER.error("angle seventh item must be bigger or equal to 0 deg.")
        assert lower2<=360, LOGGER.error("angle seventh item must be smaller or equal to 360 deg.")
        assert upper2>=0, LOGGER.error("angle eightth item must be bigger or equal to 0 deg.")
        assert upper2<=360, LOGGER.error("angle eightth item must be smaller or equal to 360 deg.")
        #lower2 *= FLOAT_TYPE( PI/FLOAT_TYPE(180.) )
        #upper2 *= FLOAT_TYPE( PI/FLOAT_TYPE(180.) )
        assert is_number(lower3), LOGGER.error("angle nineth item must be a number")
        lower3 = FLOAT_TYPE(lower3)
        assert is_number(upper3), LOGGER.error("angle tenth item must be a number")
        upper3 = FLOAT_TYPE(upper3)
        assert lower3>=0, LOGGER.error("angle nineth item must be bigger or equal to 0 deg.")
        assert lower3<=360, LOGGER.error("angle nineth item must be smaller or equal to 360 deg.")
        assert upper3>=0, LOGGER.error("angle tenth item must be bigger or equal to 0 deg.")
        assert upper3<=360, LOGGER.error("angle tenth item must be smaller or equal to 360 deg.")
        #lower3 *= FLOAT_TYPE( PI/FLOAT_TYPE(180.) )
        #upper3 *= FLOAT_TYPE( PI/FLOAT_TYPE(180.) )
        # create dihedral angles1
        if not idx1 in self.__angles:
            angles1 = {"idx2":[],"idx3":[],"idx4":[],"dihedralMap":[],"otherMap":[]}
        else:
            angles1 = {"idx2"        :self.__angles[idx1]["idx2"],
                       "idx3"        :self.__angles[idx1]["idx3"],
                       "idx4"        :self.__angles[idx1]["idx4"],
                       "dihedralMap" :self.__angles[idx1]["dihedralMap"],
                       "otherMap"    :self.__angles[idx1]["otherMap"] }
        # create dihedral angle2
        if not idx2 in self.__angles:
            angles2 = {"idx2":[],"idx3":[],"idx4":[],"dihedralMap":[],"otherMap":[]}
        else:
            angles2 = {"idx2"        :self.__angles[idx2]["idx2"],
                       "idx3"        :self.__angles[idx2]["idx3"],
                       "idx4"        :self.__angles[idx2]["idx4"],
                       "dihedralMap" :self.__angles[idx2]["dihedralMap"],
                       "otherMap"    :self.__angles[idx2]["otherMap"] }
        # create dihedral angle3
        if not idx3 in self.__angles:
            angles3 = {"idx2":[],"idx3":[],"idx4":[],"dihedralMap":[],"otherMap":[]}
        else:
            angles3 = {"idx2"        :self.__angles[idx3]["idx2"],
                       "idx3"        :self.__angles[idx3]["idx3"],
                       "idx4"        :self.__angles[idx3]["idx4"],
                       "dihedralMap" :self.__angles[idx3]["dihedralMap"],
                       "otherMap"    :self.__angles[idx3]["otherMap"] }
        # create dihedral angle4
        if not idx4 in self.__angles:
            angles4 = {"idx2":[],"idx3":[],"idx4":[],"dihedralMap":[],"otherMap":[]}
        else:
            angles4 = {"idx2"        :self.__angles[idx4]["idx2"],
                       "idx3"        :self.__angles[idx4]["idx3"],
                       "idx4"        :self.__angles[idx4]["idx4"],
                       "dihedralMap" :self.__angles[idx4]["dihedralMap"],
                       "otherMap"    :self.__angles[idx4]["otherMap"] }
        # check for re-defining
        setPos = idx2Pos = idx3Pos = idx4Pos = None
        if idx2 in angles1["idx2"] and idx3 in angles1["idx3"] and idx4 in angles1["idx4"]:
            idx2Pos = angles1["idx2"].index(idx2)
            idx3Pos = angles1["idx3"].index(idx3)
            idx4Pos = angles1["idx4"].index(idx4)
        elif idx2 in angles1["idx2"] and idx4 in angles1["idx3"] and idx3 in angles1["idx4"]:
            idx2Pos = angles1["idx2"].index(idx2)
            idx3Pos = angles1["idx4"].index(idx3)
            idx4Pos = angles1["idx3"].index(idx4)
        elif idx3 in angles1["idx2"] and idx2 in angles1["idx3"] and idx4 in angles1["idx4"]:
            idx2Pos = angles1["idx3"].index(idx2)
            idx3Pos = angles1["idx2"].index(idx3)
            idx4Pos = angles1["idx4"].index(idx4)
        elif idx4 in angles1["idx2"] and idx3 in angles1["idx3"] and idx2 in angles1["idx4"]:
            idx2Pos = angles1["idx4"].index(idx2)
            idx3Pos = angles1["idx3"].index(idx3)
            idx4Pos = angles1["idx2"].index(idx4)
        if idx2Pos is not None and (idx2Pos==idx3Pos) and (idx2Pos==idx4Pos):
            LOGGER.warn("Dihedral angle definition for atom1 index '%i' and atom2 '%i' and atom3 '%i' and atom4 '%i' is  already defined. New shells' limits [(%.3f,%.3f),(%.3f,%.3f),(%.3f,%.3f)] are set."%(idx1, idx2, idx3, idx4, lower1, upper1, lower2, upper2, lower3, upper3))
            setPos = angles1["dihedralMap"][idx2Pos]
        # set angle
        if setPos is None:
            angles1["idx2"].append(idx2)
            angles1["idx3"].append(idx3)
            angles1["idx4"].append(idx4)
            angles1["dihedralMap"].append( len(self.__anglesList[0]) )
            angles2["otherMap"].append( len(self.__anglesList[0]) )
            angles3["otherMap"].append( len(self.__anglesList[0]) )
            angles4["otherMap"].append( len(self.__anglesList[0]) )
            self.__anglesList[0] = np.append(self.__anglesList[0],idx1)
            self.__anglesList[1] = np.append(self.__anglesList[1],idx2)
            self.__anglesList[2] = np.append(self.__anglesList[2],idx3)
            self.__anglesList[3] = np.append(self.__anglesList[3],idx4)
            self.__anglesList[4] = np.append(self.__anglesList[4],lower1)
            self.__anglesList[5] = np.append(self.__anglesList[5],upper1)
            self.__anglesList[6] = np.append(self.__anglesList[6],lower2)
            self.__anglesList[7] = np.append(self.__anglesList[7],upper2)
            self.__anglesList[8] = np.append(self.__anglesList[8],lower3)
            self.__anglesList[9] = np.append(self.__anglesList[9],upper3)
        else:
            assert self.__anglesList[0][setPos] == idx1, LOGGER.error("mismatched dihedral angle atom '%s' and '%s'"%(self.__anglesList[0][setPos],idx1))
            assert sorted([idx2, idx3, idx4]) == sorted([self.__anglesList[1][idx2],self.__anglesList[2][idx3],self.__anglesList[3][idx4]]), LOGGER.error("mismatched dihedral angle atom2, atom3, atom4 at atom1 '%s'"%(idx1))
            self.__anglesList[1][setPos] = idx2
            self.__anglesList[2][setPos] = idx3
            self.__anglesList[3][setPos] = idx4
            self.__anglesList[4][setPos] = lower1
            self.__anglesList[5][setPos] = upper1
            self.__anglesList[6][setPos] = lower2
            self.__anglesList[7][setPos] = upper2
            self.__anglesList[8][setPos] = lower3
            self.__anglesList[9][setPos] = upper3
        self.__angles[idx1] = angles1
        self.__angles[idx2] = angles2
        self.__angles[idx3] = angles3
        self.__angles[idx4] = angles4
        # dump to repository
        if self.__dumpAngles:
            self._dump_to_repository({'_DihedralAngleConstraint__anglesList' :self.__anglesList,
                                      '_DihedralAngleConstraint__angles'     :self.__angles})
            # reset constraint
            self.reset_constraint()

    #@raise_if_collected
    def create_angles_by_definition(self, anglesDefinition):
        """
        Creates anglesList using angles definition.
        Calls set_angles(anglesMap) and generates angles attribute.

        :Parameters:
            #. anglesDefinition (dict): The angles definition.
               Every key must be a molecule name (residue name in pdb file).
               Every key value must be a list of angles definitions.
               Every angle definition is a list of ten items where:

               #. Name of the first dihedral atom.
               #. Name of the second dihedral atom of the first plane
                  and the first atom of the second plane.
               #. Name of the third dihedral atom of the first plane
                  and the second  atom of the second plane.
               #. Name of the fourth dihderal atom of the second plane.
               #. Minimum lower limit of the first shell or the minimum
                  angle allowed in degrees which later will be
                  converted to rad.
               #. Maximum upper limit of the first or the maximum
                  angle allowed in degrees which later will be
                  converted to rad.
               #. Minimum lower limit of the second shell or the minimum
                  angle allowed in degrees which later will be
                  converted to rad.
               #. Maximum upper limit of the second or the maximum
                  angle allowed in degrees which later will be
                  converted to rad.
               #. Minimum lower limit of the third shell or the minimum
                  angle allowed in degrees which later will be
                  converted to rad.
               #. Maximum upper limit of the third or the maximum
                  angle allowed in degrees which later will be
                  converted to rad.

        ::

            e.g. (Butane):  anglesDefinition={"BUT": [ ('C1','C2','C3','C4', 40,80, 100,140, 290,330), ] }


        """
        if self.engine is None:
            raise Exception("Engine is not defined. Can't create dihedral angles by definition")
        assert isinstance(anglesDefinition, dict), "anglesDefinition must be a dictionary"
        # check map definition
        ALL_NAMES       = self.engine.get_original_data("allNames")
        NUMBER_OF_ATOMS = self.engine.get_original_data("numberOfAtoms")
        MOLECULES_NAME  = self.engine.get_original_data("moleculesName")
        MOLECULES_INDEX = self.engine.get_original_data("moleculesIndex")
        existingMoleculesName = sorted(set(MOLECULES_NAME))
        anglesDef = {}
        for mol in anglesDefinition:
            angles = anglesDefinition[mol]
            if mol not in existingMoleculesName:
                log.LocalLogger("fullrmc").logger.warn("Molecule name '%s' in anglesDefinition is not recognized, angles definition for this particular molecule is omitted"%str(mol))
                continue
            assert isinstance(angles, (list, set, tuple)), LOGGER.error("mapDefinition molecule angles must be a list")
            angles = list(angles)
            molAnglesList = []
            for angle in angles:
                assert isinstance(angle, (list, set, tuple)), LOGGER.error("mapDefinition angles must be a list")
                angle = list(angle)
                assert len(angle)==10
                at1, at2, at3, at4, lower1, upper1, lower2, upper2, lower3, upper3 = angle
                # check for redundancy
                append = True
                for b in molAnglesList:
                    if (b[0]==at1):
                        if sorted([at2,at3,at4]) == sorted([b[1],b[2],b[3]]):
                            LOGGER.warn("Redundant definition for anglesDefinition found. The later '%s' is ignored"%str(b))
                            append = False
                            break
                if append:
                    molAnglesList.append((at1, at2, at3, at4, lower1, upper1, lower2, upper2, lower3, upper3))
            # create bondDef for molecule mol
            anglesDef[mol] = molAnglesList
        # create mols dictionary
        mols = {}
        for idx in xrange(NUMBER_OF_ATOMS):
            molName = MOLECULES_NAME[idx]
            if not molName in anglesDef:
                continue
            molIdx = MOLECULES_INDEX[idx]
            if not molIdx in mols:
                mols[molIdx] = {"name":molName, "indexes":[], "names":[]}
            mols[molIdx]["indexes"].append(idx)
            mols[molIdx]["names"].append(ALL_NAMES[idx])
        # get anglesList
        anglesList = []
        for val in mols.values():
            indexes = val["indexes"]
            names   = val["names"]
            # get definition for this molecule
            thisDef = anglesDef[val["name"]]
            for angle in thisDef:
                idx1   = indexes[ names.index(angle[0]) ]
                idx2   = indexes[ names.index(angle[1]) ]
                idx3   = indexes[ names.index(angle[2]) ]
                idx4   = indexes[ names.index(angle[3]) ]
                lower1 = angle[4]
                upper1 = angle[5]
                lower2 = angle[6]
                upper2 = angle[7]
                lower3 = angle[8]
                upper3 = angle[9]
                anglesList.append( (idx1, idx2, idx3, idx4, lower1, upper1, lower2, upper2, lower3, upper3) )
        # create angles
        self.set_angles(anglesList=anglesList)

    def compute_standard_error(self, data):
        """
        Compute the standard error (StdErr) of data not satisfying constraint conditions.

        .. math::
            StdErr = \\sum \\limits_{i}^{C}
            ( \\theta_{i} - \\theta_{i}^{min} ) ^{2}
            \\int_{0}^{\\theta_{i}^{min}} \\delta(\\theta-\\theta_{i}) d \\theta
            +
            ( \\theta_{i} - \\theta_{i}^{max} ) ^{2}
            \\int_{\\theta_{i}^{max}}^{\\pi} \\delta(\\theta-\\theta_{i}) d \\theta

        Where:\n
        :math:`C` is the total number of defined improper angles constraints. \n
        :math:`\\theta_{i}^{min}` is the improper angle constraint lower limit set for constraint i. \n
        :math:`\\theta_{i}^{max}` is the improper angle constraint upper limit set for constraint i. \n
        :math:`\\theta_{i}` is the improper angle computed for constraint i. \n
        :math:`\\delta` is the Dirac delta function. \n
        :math:`\\int_{0}^{\\theta_{i}^{min}} \\delta(\\theta-\\theta_{i}) d \\theta`
        is equal to 1 if :math:`0 \\leqslant \\theta_{i} \\leqslant \\theta_{i}^{min}` and 0 elsewhere.\n
        :math:`\\int_{\\theta_{i}^{max}}^{\\pi} \\delta(\\theta-\\theta_{i}) d \\theta`
        is equal to 1 if :math:`\\theta_{i}^{max} \\leqslant \\theta_{i} \\leqslant \\pi` and 0 elsewhere.\n

        :Parameters:
            #. data (numpy.array): The constraint value data to compute standardError.

        :Returns:
            #. standardError (number): The calculated standardError of the constraint.
        """
        return FLOAT_TYPE( np.sum(data["reducedAngles"]**2) )

    def get_constraint_value(self):
        """
        Get constraint's data.

        :Returns:
            #. data (numpy.array): The constraint value data
        """
        return self.data

    @reset_if_collected_out_of_date
    def compute_data(self):
        """ Compute constraint's data."""
        if len(self._atomsCollector):
            anglesData    = np.zeros(self.__anglesList[0].shape[0], dtype=FLOAT_TYPE)
            reducedData   = np.zeros(self.__anglesList[0].shape[0], dtype=FLOAT_TYPE)
            anglesIndexes = set(set(range(self.__anglesList[0].shape[0])))
            anglesIndexes  = list( anglesIndexes-self._atomsCollector._randomData )

            indexes1 = self._atomsCollector.get_relative_indexes(self.__anglesList[0][anglesIndexes])
            indexes2 = self._atomsCollector.get_relative_indexes(self.__anglesList[1][anglesIndexes])
            indexes3 = self._atomsCollector.get_relative_indexes(self.__anglesList[2][anglesIndexes])
            indexes4 = self._atomsCollector.get_relative_indexes(self.__anglesList[3][anglesIndexes])
            lowerLimit1 = self.__anglesList[4][anglesIndexes]
            upperLimit1 = self.__anglesList[5][anglesIndexes]
            lowerLimit2 = self.__anglesList[6][anglesIndexes]
            upperLimit2 = self.__anglesList[7][anglesIndexes]
            lowerLimit3 = self.__anglesList[8][anglesIndexes]
            upperLimit3 = self.__anglesList[9][anglesIndexes]
        else:
            indexes1 = self._atomsCollector.get_relative_indexes(self.__anglesList[0])
            indexes2 = self._atomsCollector.get_relative_indexes(self.__anglesList[1])
            indexes3 = self._atomsCollector.get_relative_indexes(self.__anglesList[2])
            indexes4 = self._atomsCollector.get_relative_indexes(self.__anglesList[3])
            lowerLimit1 = self.__anglesList[4]
            upperLimit1 = self.__anglesList[5]
            lowerLimit2 = self.__anglesList[6]
            upperLimit2 = self.__anglesList[7]
            lowerLimit3 = self.__anglesList[8]
            upperLimit3 = self.__anglesList[9]
        # compute data
        angles, reduced =  full_dihedral_angles_coords( indexes1           = indexes1,
                                                        indexes2           = indexes2,
                                                        indexes3           = indexes3,
                                                        indexes4           = indexes4,
                                                        lowerLimit1        = lowerLimit1,
                                                        upperLimit1        = upperLimit1,
                                                        lowerLimit2        = lowerLimit2,
                                                        upperLimit2        = upperLimit2,
                                                        lowerLimit3        = lowerLimit3,
                                                        upperLimit3        = upperLimit3,
                                                        boxCoords          = self.engine.boxCoordinates,
                                                        basis              = self.engine.basisVectors ,
                                                        isPBC              = self.engine.isPBC,
                                                        reduceAngleToUpper = False,
                                                        reduceAngleToLower = False,
                                                        ncores             = INT_TYPE(1))
        # create full length data
        if len(self._atomsCollector):
            anglesData[anglesIndexes]  = angles
            reducedData[anglesIndexes] = reduced
            angles  = anglesData
            reduced = reducedData
        # set data.
        self.set_data( {"angles":angles, "reducedAngles":reduced} )
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)
        # set standardError
        self.set_standard_error( self.compute_standard_error(data = self.data) )
        # set original data
        if self.originalData is None:
            self._set_original_data(self.data)

    def compute_before_move(self, realIndexes, relativeIndexes):
        """
        Compute constraint's data before move is executed.

        :Parameters:
            #. realIndexes (numpy.ndarray): Group atoms index the move will
               be applied to.
            #. relativeIndexes (numpy.ndarray): Not used here.
        """
        # get angles indexes
        anglesIndexes = []
        #for idx in relativeIndexes:
        for idx in realIndexes:
            anglesIndexes.extend( self.__angles[idx]['dihedralMap'] )
            anglesIndexes.extend( self.__angles[idx]['otherMap'] )
        #anglesIndexes = list(set(anglesIndexes))
        anglesIndexes = list( set(anglesIndexes)-set(self._atomsCollector._randomData) )
        # compute data before move
        if len(anglesIndexes):
            angles, reduced =  full_dihedral_angles_coords( indexes1           = self._atomsCollector.get_relative_indexes(self.__anglesList[0][anglesIndexes]),
                                                            indexes2           = self._atomsCollector.get_relative_indexes(self.__anglesList[1][anglesIndexes]),
                                                            indexes3           = self._atomsCollector.get_relative_indexes(self.__anglesList[2][anglesIndexes]),
                                                            indexes4           = self._atomsCollector.get_relative_indexes(self.__anglesList[3][anglesIndexes]),
                                                            lowerLimit1        = self.__anglesList[4][anglesIndexes],
                                                            upperLimit1        = self.__anglesList[5][anglesIndexes],
                                                            lowerLimit2        = self.__anglesList[6][anglesIndexes],
                                                            upperLimit2        = self.__anglesList[7][anglesIndexes],
                                                            lowerLimit3        = self.__anglesList[8][anglesIndexes],
                                                            upperLimit3        = self.__anglesList[9][anglesIndexes],
                                                            boxCoords          = self.engine.boxCoordinates,
                                                            basis              = self.engine.basisVectors ,
                                                            isPBC              = self.engine.isPBC,
                                                            reduceAngleToUpper = False,
                                                            reduceAngleToLower = False,
                                                            ncores             = INT_TYPE(1))
        else:
            angles  = None
            reduced = None
        # set data before move
        self.set_active_atoms_data_before_move( {"anglesIndexes":anglesIndexes, "angles":angles, "reducedAngles":reduced} )
        self.set_active_atoms_data_after_move(None)

    def compute_after_move(self, realIndexes, relativeIndexes, movedBoxCoordinates):
        """
        Compute constraint's data after move is executed.

        :Parameters:
            #. realIndexes (numpy.ndarray): Not used here.
            #. relativeIndexes (numpy.ndarray): Group atoms relative index
               the move will be applied to.
            #. movedBoxCoordinates (numpy.ndarray): The moved atoms new
               coordinates.
        """
        # get angles indexes
        anglesIndexes = self.activeAtomsDataBeforeMove["anglesIndexes"]
        # change coordinates temporarily
        boxData = np.array(self.engine.boxCoordinates[relativeIndexes], dtype=FLOAT_TYPE)
        self.engine.boxCoordinates[relativeIndexes] = movedBoxCoordinates
        # compute data before move
        if len(anglesIndexes):
            angles, reduced =  full_dihedral_angles_coords( indexes1           = self._atomsCollector.get_relative_indexes(self.__anglesList[0][anglesIndexes]),
                                                            indexes2           = self._atomsCollector.get_relative_indexes(self.__anglesList[1][anglesIndexes]),
                                                            indexes3           = self._atomsCollector.get_relative_indexes(self.__anglesList[2][anglesIndexes]),
                                                            indexes4           = self._atomsCollector.get_relative_indexes(self.__anglesList[3][anglesIndexes]),
                                                            lowerLimit1        = self.__anglesList[4][anglesIndexes],
                                                            upperLimit1        = self.__anglesList[5][anglesIndexes],
                                                            lowerLimit2        = self.__anglesList[6][anglesIndexes],
                                                            upperLimit2        = self.__anglesList[7][anglesIndexes],
                                                            lowerLimit3        = self.__anglesList[8][anglesIndexes],
                                                            upperLimit3        = self.__anglesList[9][anglesIndexes],
                                                            boxCoords          = self.engine.boxCoordinates,
                                                            basis              = self.engine.basisVectors ,
                                                            isPBC              = self.engine.isPBC,
                                                            reduceAngleToUpper = False,
                                                            reduceAngleToLower = False,
                                                            ncores             = INT_TYPE(1))
        else:
            angles  = None
            reduced = None
        # set active data after move
        self.set_active_atoms_data_after_move( {"angles":angles, "reducedAngles":reduced} )
        # reset coordinates
        self.engine.boxCoordinates[relativeIndexes] = boxData
        # compute standardError after move
        if angles is None:
            self.set_after_move_standard_error( self.standardError )
        else:
            # anglesIndexes is a fancy slicing, RL is a copy not a view.
            RL = self.data["reducedAngles"][anglesIndexes]
            self.data["reducedAngles"][anglesIndexes] += reduced-self.activeAtomsDataBeforeMove["reducedAngles"]
            self.set_after_move_standard_error( self.compute_standard_error(data = self.data) )
            self.data["reducedAngles"][anglesIndexes] = RL

    def accept_move(self, realIndexes, relativeIndexes):
        """
        Accept move.

        #. realIndexes (numpy.ndarray): Not used here.
        #. relativeIndexes (numpy.ndarray): Not used here.
        """
        # get indexes
        anglesIndexes = self.activeAtomsDataBeforeMove["anglesIndexes"]
        if len(anglesIndexes):
            # set new data
            data = self.data
            data["angles"][anglesIndexes]        += self.activeAtomsDataAfterMove["angles"]-self.activeAtomsDataBeforeMove["angles"]
            data["reducedAngles"][anglesIndexes] += self.activeAtomsDataAfterMove["reducedAngles"]-self.activeAtomsDataBeforeMove["reducedAngles"]
            self.set_data( data )
            # update standardError
            self.set_standard_error( self.afterMoveStandardError )
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)

    def reject_move(self, realIndexes, relativeIndexes):
        """
        Reject move

        :Parameters:
            #. realIndexes (numpy.ndarray): Not used here.
            #. relativeIndexes (numpy.ndarray): Not used here.
        """
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)
        self.set_active_atoms_data_after_move(None)

    def accept_amputation(self, realIndex, relativeIndex):
        """
        Accept amputation of atom and set constraint's data and standard
        error accordingly.

        :Parameters:
            #. realIndex (numpy.ndarray): Atom's index as a numpy array
               of a single element.
            #. relativeIndex (numpy.ndarray): Not used here.
        """
        # MAYBE WE DON"T NEED TO CHANGE DATA AND SE. BECAUSE THIS MIGHT BE A PROBLEM
        # WHEN IMPLEMENTING ATOMS RELEASING. MAYBE WE NEED TO COLLECT DATA INSTEAD, REMOVE
        # AND ADD UPON RELEASE
        # get all involved data
        anglesIndexes = []
        for idx in realIndex:
            anglesIndexes.extend( self.__angles[idx]['dihedralMap'] )
            anglesIndexes.extend( self.__angles[idx]['otherMap'] )
        anglesIndexes = list(set(anglesIndexes))
        if len(anglesIndexes):
            # set new data
            data = self.data
            data["angles"][anglesIndexes]        = 0
            data["reducedAngles"][anglesIndexes] = 0
            self.set_data( data )
            # update standardError
            SE = self.compute_standard_error(data = self.get_constraint_value())
            self.set_standard_error( SE )
        # reset activeAtoms data
        self.set_active_atoms_data_before_move(None)

    def reject_amputation(self, realIndex, relativeIndex):
        """
        Reject amputation of atom.

        :Parameters:
            #. realIndex (numpy.ndarray): Not used here.
            #. relativeIndex (numpy.ndarray): Not used here.
        """
        pass

    def _on_collector_collect_atom(self, realIndex):
        # get angle indexes
        AI = self.__angles[realIndex]['dihedralMap'] + self.__angles[realIndex]['otherMap']
        # append all mapped bonds to collector's random data
        self._atomsCollector._randomData = self._atomsCollector._randomData.union( set(AI) )
        # collect atom anglesIndexes
        self._atomsCollector.collect(realIndex, dataDict={'dihedralMap':self.__angles[realIndex]['dihedralMap'],
                                                          'otherMap'   :self.__angles[realIndex]['otherMap']})


    def plot(self, ax=None, nbins=20, subplots=True, split=None,
                   wspace=0.3, hspace=0.3,
                   histTtype='bar', lineWidth=None, lineColor=None,
                   xlabel=True, xlabelSize=16,
                   ylabel=True, ylabelSize=16,
                   legend=True, legendCols=1, legendLoc='best',
                   title=True, titleStdErr=True, titleAtRem=True,
                   titleUsedFrame=True, show=True):
        """
        Plot dihedral angles constraint's distribution histogram.

        :Parameters:
            #. ax (None, matplotlib Axes): matplotlib Axes instance to plot in.
               If ax is given,  subplots parameters will be omitted.
               If None is given a new plot figure will be created.
            #. nbins (int): number of bins in histogram.
            #. subplots (boolean): Whether to add plot constraint on multiple
               axes.
            #. split (None, 'name', 'element'): To split plots into histogram
               per atom names, elements in addition to lower and upper bounds.
               If None is given, histograms will be built from lower and upper
               bounds only.
            #. wspace (float): The amount of width reserved for blank space
               between subplots, expressed as a fraction of the average axis
               width.
            #. hspace (float): The amount of height reserved for white space
               between subplots, expressed as a fraction of the average axis
               height.
            #. histTtype (string): the histogram type. optional among
               ['bar', 'barstacked', 'step', 'stepfilled']
            #. lineWidth (None, integer): bars contour line width.
               If None is given, then default value set automatically.
            #. lineColor (None, integer): bars contour line color.
               If None is given, then default value set automatically.
            #. xlabel (boolean): Whether to create x label.
            #. xlabelSize (number): The x label font size.
            #. ylabel (boolean): Whether to create y label.
            #. ylabelSize (number): The y label font size.
            #. legend (boolean): Whether to create the legend or not
            #. legendCols (integer): Legend number of columns.
            #. legendLoc (string): The legend location. Anything among
               'right', 'center left', 'upper right', 'lower right', 'best',
               'center', 'lower left', 'center right', 'upper left',
               'upper center', 'lower center' is accepted.
            #. title (boolean): Whether to create the title or not.
            #. titleStdErr (boolean): Whether to show constraint standard
               error value in title.
            #. titleAtRem (boolean): Whether to show engine's number of
               removed atoms.
            #. titleUsedFrame(boolean): Whether to show used frame name in
               title.
            #. show (boolean): Whether to render and show figure before
               returning.

        :Returns:
            #. figure (matplotlib Figure): matplotlib used figure.
            #. axes (matplotlib Axes, List): matplotlib axes or a list of axes.
        """
        def _get_bins(dmin, dmax, boundaries, nbins):
            # create bins
            delta = float(dmax-dmin)/float(nbins-1)
            bins  = range(nbins)
            bins  = [b*delta for b in bins]
            bins  = [b+dmin for b in bins]
            # check boundaries
            bidx = 0
            for b in sorted(boundaries):
                for i in range(bidx, len(bins)-1):
                    bidx = i
                    # exact match with boundary
                    if b==bins[bidx]:
                        break
                    # boundary between two bins, move closest bin to boundary
                    if bins[bidx] < b < bins[bidx+1]:
                        if b-bins[bidx] > bins[bidx+1]-b:
                            bins[bidx+1] = b
                        else:
                            bins[bidx]   = b
                        break
            # return bins
            return bins
        # get constraint value
        output = self.get_constraint_value()
        if output is None:
            LOGGER.warn("%s constraint data are not computed."%(self.__class__.__name__))
            return
       # compute categories
        if split == 'name':
            splitV = self.engine.get_original_data("allNames")
        elif split == 'element':
            splitV = self.engine.get_original_data("allElements")
        else:
            splitV = None
        categories = {}
        atom2  = self.__anglesList[0]
        atom1  = self.__anglesList[1]
        atom3  = self.__anglesList[2]
        atom4  = self.__anglesList[3]
        lower1 = self.__anglesList[4]
        upper1 = self.__anglesList[5]
        lower2 = self.__anglesList[6]
        upper2 = self.__anglesList[7]
        lower3 = self.__anglesList[8]
        upper3 = self.__anglesList[9]
        for idx in xrange(self.__anglesList[0].shape[0]):
            if self._atomsCollector.is_collected(idx):
                continue
            if splitV is not None:
                a1 = splitV[ atom1[idx] ]
                a2 = splitV[ atom2[idx] ]
                a3 = splitV[ atom3[idx] ]
                a4 = splitV[ atom4[idx] ]
            else:
                a1 = a2 = a3 = a4 = ''
            l1 = lower1[idx]
            u1 = upper1[idx]
            l2 = lower2[idx]
            u2 = upper2[idx]
            l3 = lower3[idx]
            u3 = upper3[idx]
            k = (a1,a2,a3,a4,l1,u1,l2,u2,l3,u3)
            L = categories.get(k, [])
            L.append(idx)
            categories[k] = L
        ncategories = len(categories)
        # import matplotlib
        import matplotlib.pyplot as plt
        # get axes
        if ax is None:
            if subplots and ncategories>1:
                x = np.ceil(np.sqrt(ncategories))
                y = np.ceil(ncategories/x)
                FIG, N_AXES = plt.subplots(int(x), int(y) )
                N_AXES = FIG.flatten()
                FIG.subplots_adjust(wspace=wspace, hspace=hspace)
                [N_AXES[i].axis('off') for i in range(ncategories,len(N_AXES))]
            else:
                FIG  = plt.figure()
                AXES = FIG.gca()
                subplots = False
        else:
            AXES = ax
            FIG = AXES.get_figure()
            subplots = False
        # start plotting
        COLORS = ["b",'g','r','c','y','m']
        if subplots:
            for idx, key in enumerate(categories):
                a1,a2,a3,a4, L1,U1, L2,U2, L3,U3  = key
                LU = sorted(set( [(L1,U1),(L2,U2),(L3,U3)] ))
                LA = " ".join( ["(%.2f,%.2f)"%(l,u)  for l,u in LU] )
                label = "%s%s%s%s%s%s%s%s"%(a1,'-'*(len(a1)>0),a2,'-'*(len(a1)>0),a3,'-'*(len(a1)>0),a4,LA)
                COL  = COLORS[idx%len(COLORS)]
                AXES = N_AXES[idx]
                idxs = categories[key]
                data = self.data["angles"][idxs]
                # get data limits
                mn = np.min(data)
                mx = np.max(data)
                # get bins
                BINS = _get_bins(dmin=mn, dmax=mx, boundaries=[L1,U1,L2,U2,L3,U3], nbins=nbins)
                # plot histogram
                D, _, P = AXES.hist(x=data, bins=BINS,
                                    color=COL, label=label,
                                    histtype=histTtype)
                # vertical lines
                Y = max(D)
                for idx, (l,u) in enumerate(LU):
                    AXES.plot([l,l],[0,Y+0.1*Y], linewidth=1.0, color='k', linestyle=['--','-.',':'][idx])
                    AXES.plot([u,u],[0,Y+0.1*Y], linewidth=1.0, color='k', linestyle=['--','-.',':'][idx])
                # legend
                if legend:
                    AXES.legend(frameon=False, ncol=legendCols, loc=legendLoc)
                # set axis labels
                if xlabel:
                    AXES.set_xlabel("$deg.$", size=xlabelSize)
                if ylabel:
                    AXES.set_ylabel("$number$"  , size=ylabelSize)
                if lineWidth is not None:
                    [p.set_linewidth(lineWidth) for p in P]
                if lineColor is not None:
                    [p.set_edgecolor(lineColor) for p in P]
                # update limits
                AXES.set_xmargin(0.1)
                AXES.autoscale()
        else:
            for idx, key in enumerate(categories):
                a1,a2,a3,a4, L1,U1, L2,U2, L3,U3  = key
                LU = sorted(set( [(L1,U1),(L2,U2),(L3,U3)] ))
                LA = " ".join( ["(%.2f,%.2f)"%(l,u)  for l,u in LU] )
                label = "%s%s%s%s%s%s%s%s"%(a1,'-'*(len(a1)>0),a2,'-'*(len(a1)>0),a3,'-'*(len(a1)>0),a4,LA)
                COL  = COLORS[idx%len(COLORS)]
                idxs = categories[key]
                data = self.data["angles"][idxs]
                # get data limits
                mn = np.min(data)
                mx = np.max(data)
                # get bins
                BINS = _get_bins(dmin=mn, dmax=mx, boundaries=[L1,U1,L2,U2,L3,U3], nbins=nbins)
                # plot histogram
                D, _, P = AXES.hist(x=data, bins=BINS,
                                    color=COL, label=label,
                                    histtype=histTtype)
                # vertical lines
                Y = max(D)
                for idx, (l,u) in enumerate(LU):
                    AXES.plot([l,l],[0,Y+0.1*Y], linewidth=1.0, color='k', linestyle=['--','-.',':'][idx])
                    AXES.plot([u,u],[0,Y+0.1*Y], linewidth=1.0, color='k', linestyle=['--','-.',':'][idx])
                if lineWidth is not None:
                    [p.set_linewidth(lineWidth) for p in P]
                if lineColor is not None:
                    [p.set_edgecolor(lineColor) for p in P]
            # update limits
            AXES.set_xmargin(0.1)
            AXES.autoscale()
            # legend
            if legend:
                AXES.legend(frameon=False, ncol=legendCols, loc=legendLoc)
            # set axis labels
            if xlabel:
                AXES.set_xlabel("$deg.$", size=xlabelSize)
            if ylabel:
                AXES.set_ylabel("$number$"  , size=ylabelSize)
        # set title
        if title:
            FIG.canvas.set_window_title('Dihedral Angle Constraint')
            if titleUsedFrame:
                t = '$frame: %s$ : '%self.engine.usedFrame.replace('_','\\_')
            else:
                t = ''
            if titleAtRem:
                t += "$%i$ $rem.$ $at.$ - "%(len(self.engine._atomsCollector))
            if titleStdErr and self.standardError is not None:
                t += "$std$ $error=%.6f$ "%(self.standardError)
            if len(t):
                FIG.suptitle(t, fontsize=14)

        # set background color
        FIG.patch.set_facecolor('white')
        #show
        if show:
            plt.show()
        # return axes
        if subplots:
            return FIG, N_AXES
        else:
            return FIG, AXES

    def export(self, fname, delimiter='     ', comments='# ', split=None):
        """
        Export dihedral angles constraint's distribution histogram.

        :Parameters:
            #. fname (path): full file name and path.
            #. delimiter (string): String or character separating columns.
            #. comments (string): String that will be prepended to the header.
            #. split (None, 'name', 'element'): To split output into per atom
               names, elements in addition to lower and upper bounds.
               If None is given, output will be built from lower and upper
               bounds only.
        """
        # get constraint value
        output = self.get_constraint_value()
        if output is None:
            LOGGER.warn("%s constraint data are not computed."%(self.__class__.__name__))
            return
        # compute categories
        if split == 'name':
            splitV = self.engine.get_original_data("allNames")
        elif split == 'element':
            splitV = self.engine.get_original_data("allElements")
        else:
            splitV = None
        categories = {}
        atom2  = self.__anglesList[0]
        atom1  = self.__anglesList[1]
        atom3  = self.__anglesList[2]
        atom4  = self.__anglesList[3]
        lower1 = self.__anglesList[4]
        upper1 = self.__anglesList[5]
        lower2 = self.__anglesList[6]
        upper2 = self.__anglesList[7]
        lower3 = self.__anglesList[8]
        upper3 = self.__anglesList[9]
        for idx in xrange(self.__anglesList[0].shape[0]):
            if self._atomsCollector.is_collected(idx):
                continue
            if splitV is not None:
                a1 = splitV[ atom1[idx] ]
                a2 = splitV[ atom2[idx] ]
                a3 = splitV[ atom3[idx] ]
                a4 = splitV[ atom4[idx] ]
            else:
                a1 = a2 = a3 = a4 = ''
            l1 = lower1[idx]
            u1 = upper1[idx]
            l2 = lower2[idx]
            u2 = upper2[idx]
            l3 = lower3[idx]
            u3 = upper3[idx]
            k = (a1,a2,a3,a4,l1,u1,l2,u2,l3,u3)
            L = categories.get(k, [])
            L.append(idx)
            categories[k] = L
        ncategories = len(categories)
        # create data
        for idx, key in enumerate(categories):
            idxs = categories[key]
            data = self.data["angles"][idxs]
            categories[key] = [str(d) for d in data]
        # adjust data size
        maxSize = max( [len(v) for v in categories.values()] )
        for key in categories:
            data = categories[key]
            add  = maxSize-len(data)
            if add > 0:
                categories[key] = data + ['']*add
        # start creating header and data
        sortCa = sorted( categories )
        header = []
        for key in sortCa:
            a1,a2,a3,a4, L1,U1, L2,U2, L3,U3  = key
            LU = sorted(set( [(L1,U1),(L2,U2),(L3,U3)] ))
            LA = " ".join( ["(%.2f,%.2f)"%(l,u)  for l,u in LU] )
            header.append( ("%s%s%s%s%s%s%s%s"%(a1,'-'*(len(a1)>0),a2,'-'*(len(a1)>0),a3,'-'*(len(a1)>0),a4,LA)).replace(' ','') )
        data   = [categories[key] for key in sortCa]
        # save
        data = np.transpose(data)
        np.savetxt(fname     = fname,
                   X         = data,
                   fmt       = '%s',
                   delimiter = delimiter,
                   header    = " ".join(header),
                   comments  = comments)



##########################################################################################
##############################  IMPORTING USEFUL DEFINITIONS  ############################
# standard libraries imports
from __future__ import print_function
import os, sys

# external libraries imports
import numpy as np

# fullrmc library imports
from fullrmc.Globals import LOGGER, maxint
from fullrmc.Engine import Engine
from fullrmc.Generators.Agitations import DistanceAgitationGenerator, AngleAgitationGenerator
from fullrmc.Selectors.OrderedSelectors import DefinedOrderSelector
from fullrmc.Constraints.BondConstraints import BondConstraint
from fullrmc.Constraints.AngleConstraints import BondsAngleConstraint


##########################################################################################
##################################  SHUT DOWN LOGGING  ###################################
LOGGER.set_minimum_level(maxint, stdoutFlag=True, fileFlag=True)


##########################################################################################
#####################################  CREATE ENGINE  ####################################
pdbPath = "waterBox.pdb"
ENGINE = Engine(path=None)
ENGINE.set_pdb( pdbPath  )

# add constraints
B_CONSTRAINT  = BondConstraint()
BA_CONSTRAINT = BondsAngleConstraint()
ENGINE.add_constraints([B_CONSTRAINT, BA_CONSTRAINT])
B_CONSTRAINT.create_bonds_by_definition( bondsDefinition={"TIP": [('OH2' ,'H1' , 0.8, 1.1),
                                                                  ('OH2' ,'H2' , 0.8, 1.1)] })
BA_CONSTRAINT.create_angles_by_definition( anglesDefinition={"TIP": [ ('OH2'  ,'H1' ,'H2' , 85, 120)] })

##########################################################################################
####################################  DIFFERENT RUNS  ####################################
def agitate_bonds():
    print("Agitate bonds")
    # agitate bonds
    groups = []
    for idx in range(0,ENGINE.pdb.numberOfAtoms,3):
        groups.append(np.array([idx, idx+1]))
        groups.append(np.array([idx, idx+2]))
    ENGINE.set_groups(groups)
    [g.set_move_generator(DistanceAgitationGenerator(amplitude=0.1,agitate=(False,True))) for g in ENGINE.groups]
    # set ordered selector
    ENGINE.set_group_selector(DefinedOrderSelector(ENGINE, order=None) )
    # run engine
    xyzPath="bonds.xyz"
    if os.path.isfile(xyzPath): os.remove(xyzPath)
    nsteps = 250*len(ENGINE.groups)
    xyzFrequency = len(ENGINE.groups)
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=2*nsteps, xyzFrequency=xyzFrequency, xyzPath=xyzPath, restartPdb=None)

def agitate_angles():
    print("Agitate angles")
    # agitate angles
    ENGINE.set_groups_as_molecules()
    [g.set_move_generator(AngleAgitationGenerator(amplitude=5)) for g in ENGINE.groups]
    # set ordered selector
    ENGINE.set_group_selector(DefinedOrderSelector(ENGINE, order=None) )
    # run engine
    xyzPath="angles.xyz"
    if os.path.isfile(xyzPath): os.remove(xyzPath)
    nsteps = 250*len(ENGINE.groups)
    xyzFrequency = len(ENGINE.groups)
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=2*nsteps, xyzFrequency=xyzFrequency, xyzPath=xyzPath, restartPdb=None)

def agitate_both():
    print("Agitate both")
    # agitate all
    ENGINE.set_groups_as_molecules()
    groups = [g for g in ENGINE.groups]
    for idx in range(0,ENGINE.pdb.numberOfAtoms,3):
        groups.append(np.array([idx, idx+1]))
        groups.append(np.array([idx, idx+2]))
    ENGINE.add_groups(groups)
    # set move generator
    for g in ENGINE.groups:
        if len(g)==2:
            g.set_move_generator(DistanceAgitationGenerator(amplitude=0.1, agitate=(False,True)))
        elif len(g)==3:
            g.set_move_generator(AngleAgitationGenerator(amplitude=5))
        else:
            raise
    # set ordered selector
    ENGINE.set_group_selector(DefinedOrderSelector(ENGINE, order=None) )
    # set ordered selector
    ENGINE.set_group_selector(DefinedOrderSelector(ENGINE, order=None) )
    # run engine
    xyzPath="both.xyz"
    if os.path.isfile(xyzPath): os.remove(xyzPath)
    nsteps = 250*len(ENGINE.groups)
    xyzFrequency = len(ENGINE.groups)
    # run engine
    ENGINE.run(numberOfSteps=nsteps, saveFrequency=2*nsteps, xyzFrequency=xyzFrequency, xyzPath=xyzPath, restartPdb=None)


##########################################################################################
#####################################  RUN SIMULATION  ###################################
agitate_bonds()
agitate_angles()
agitate_both()


##########################################################################################
#################################  VISUALIZE SIMULATION  #################################
ENGINE.set_pdb(pdbPath)
ENGINE.visualize( commands = ["bonds.xyz", "angles.xyz", "both.xyz"],
                  boxToCenter=True, boxWidth=1,
                  representationParams='CPK 1.0 0.2 50 50')



##########################################################################################
#############################  IMPORTING USEFUL DEFINITIONS  #############################
# standard distribution imports
import os, sys

# external libraries imports

# fullrmc library imports
from fullrmc.Globals import LOGGER
from fullrmc.Engine import Engine
from fullrmc.Constraints.PairCorrelationConstraints import PairDistributionConstraint
from fullrmc.Constraints.DistanceConstraints import InterMolecularDistanceConstraint
from fullrmc.Constraints.BondConstraints import BondConstraint
from fullrmc.Constraints.AngleConstraints import BondsAngleConstraint
from fullrmc.Core.GroupSelector import RecursiveGroupSelector
from fullrmc.Selectors.RandomSelectors import RandomSelector
from fullrmc.Core.MoveGenerator import MoveGeneratorCollector
from fullrmc.Generators.Translations import TranslationGenerator
from fullrmc.Generators.Rotations import RotationGenerator

##########################################################################################
#####################################  CREATE ENGINE  ####################################
try:
    DIR_PATH = os.path.dirname( os.path.realpath(__file__) )
except:
    DIR_PATH = ''

# engine file names
engineFileName = "CO2.rmc"
expFileName    = "Xrays.gr"
pdbFileName    = "CO2.pdb"
freshStart     = False

# engine variables
expDataPath = os.path.join(DIR_PATH, expFileName)
pdbPath     = os.path.join(DIR_PATH, pdbFileName)
enginePath  = os.path.join(DIR_PATH, engineFileName)
FRESH_START = False

ENGINE = Engine(path=None)
if not ENGINE.is_engine(enginePath) or FRESH_START:
    # initialize engine
    ENGINE = Engine(path=enginePath, freshStart=True)
    ENGINE.set_pdb(pdbPath)
    PDF_CONSTRAINT = PairDistributionConstraint(experimentalData=expDataPath, weighting="atomicNumber")
    IMD_CONSTRAINT = InterMolecularDistanceConstraint(defaultDistance=1.4)
    B_CONSTRAINT   = BondConstraint()
    BA_CONSTRAINT  = BondsAngleConstraint()
    # add constraints
    ENGINE.add_constraints([PDF_CONSTRAINT, IMD_CONSTRAINT, B_CONSTRAINT, BA_CONSTRAINT])
    B_CONSTRAINT.create_bonds_by_definition( bondsDefinition={"CO2": [('C' ,'O1' , 0.52, 1.4),
                                                                      ('C' ,'O2' , 0.52, 1.4)] })
    BA_CONSTRAINT.create_angles_by_definition( anglesDefinition={"CO2": [ ('C' ,'O1' ,'O2' , 170, 180)] })
    # initialize constraints data
    ENGINE.initialize_used_constraints()
    # save engine
    ENGINE.save()
else:
    ENGINE = ENGINE.load(enginePath)
    # unpack constraints before fitting in case tweaking is needed
    PDF_CONSTRAINT, IMD_CONSTRAINT, B_CONSTRAINT, BA_CONSTRAINT = ENGINE.constraints


# set all constraints as used. Used value is True by default!
# Now you know you can deactivate any constraint at any time though.
PDF_CONSTRAINT.set_used(True)
IMD_CONSTRAINT.set_used(True)
B_CONSTRAINT.set_used(True)
BA_CONSTRAINT.set_used(True)

##########################################################################################
#####################################  DIFFERENT RUNS  ###################################
def run_atoms(ENGINE, rang=None, recur=None, xyzFrequency=500):
    ENGINE.set_groups(None)
    # set selector
    if recur is None: recur = 10
    ENGINE.set_group_selector(RandomSelector(ENGINE))
    # number of steps
    nsteps = recur*len(ENGINE.groups)
    if rang is None: rang=20
    for stepIdx in range(rang):
        LOGGER.info("Running 'atoms' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps, xyzFrequency=xyzFrequency, xyzPath="atomsTraj.xyz", restartPdb=None)

def run_molecules(ENGINE, rang=5, recur=100, refine=False, explore=True, exportPdb=False, xyzFrequency=500):
    ENGINE.set_groups_as_molecules()
    [g.set_move_generator( MoveGeneratorCollector(collection=[TranslationGenerator(amplitude=0.2),RotationGenerator(amplitude=2)],randomize=True) ) for g in ENGINE.groups]
    # number of steps
    nsteps = 20*len(ENGINE.groups)
    # set selector
    gs = RecursiveGroupSelector(RandomSelector(ENGINE), recur=recur, refine=refine, explore=explore)
    ENGINE.set_group_selector(gs)
    for stepIdx in range(rang):
        LOGGER.info("Running 'molecules' mode step %i"%(stepIdx))
        ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps, xyzFrequency=xyzFrequency, xyzPath="moleculeTraj.xyz", restartPdb=None)

def run_recurring_atoms(ENGINE, rang=None, recur=None, explore=True, refine=False, xyzFrequency=500):
    ENGINE.set_groups(None)
    # set selector
    if recur is None: recur = 10
    gs = RecursiveGroupSelector(RandomSelector(ENGINE), recur=recur, refine=refine, explore=explore)
    ENGINE.set_group_selector(gs)
    # number of steps
    nsteps = recur*len(ENGINE.groups)
    if rang is None: rang=20
    for stepIdx in range(rang):
        LOGGER.info("Running 'explore' mode step %i"%(stepIdx))
        if explore:
            ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps, xyzFrequency=xyzFrequency, xyzPath="exploreTraj.xyz", restartPdb=None)
        elif refine:
            ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps, xyzFrequency=xyzFrequency, xyzPath="refineTraj.xyz", restartPdb=None)
        else:
            ENGINE.run(numberOfSteps=nsteps, saveFrequency=nsteps, xyzFrequency=xyzFrequency, xyzPath="recurTraj.xyz", restartPdb=None)

##########################################################################################
####################################  RUN SIMULATION  ####################################
## remove all .xyz trajectory files
files = [f for f in os.listdir(".") if os.path.isfile(f) and ".xyz" in f]
[os.remove(fname) for fname in files]
## run atoms
run_atoms(ENGINE, rang=4, xyzFrequency=None)
#run_molecules(ENGINE, xyzFrequency=None)
run_recurring_atoms(ENGINE, rang=50, explore=True, refine=False, xyzFrequency=None)
run_recurring_atoms(ENGINE, rang=4, explore=False, refine=True, xyzFrequency=None)
run_atoms(ENGINE, rang=10, xyzFrequency=None)


##########################################################################################
###################################### CALL plot.py ######################################
os.system("%s %s"%(sys.executable, os.path.join(DIR_PATH, 'plot.py')))



import random, math
import numpy as np
import matplotlib.pyplot as plt
from pdbparser import pdbparser
from pdbparser.Utilities import Geometry, Information, Modify

## read pdb file
PDB = pdbparser("SiOx.pdb")

## translate to origin
COM = Geometry.get_center(PDB.indexes, PDB)
Geometry.translate(PDB.indexes, PDB, -COM)

## get radius
minX,maxX, minY,maxY, minZ,maxZ = Geometry.get_min_max(PDB.indexes, PDB)
radiusX = (maxX-minX)/2.
radiusY = (maxY-minY)/2.
radiusZ = (maxZ-minZ)/2.
radius  = max(radiusX,radiusY,radiusZ)

# get oxygen/silicon ratio vs depth
oIdxs   = Information.get_records_indexes_by_attribute_value(PDB.indexes, PDB, 'element_symbol', 'O')
siIdxs  = Information.get_records_indexes_by_attribute_value(PDB.indexes, PDB, 'element_symbol', 'Si')
oDist   = np.sqrt(np.sum(PDB.coordinates[oIdxs]**2,1))
siDist  = np.sqrt(np.sum(PDB.coordinates[siIdxs]**2,1))
depth   = math.ceil(max(oDist) - min(oDist))
# get histogram of oxygen/silicon distribution ratio
ho,  e = np.histogram(oDist,  bins=range(math.ceil(radius) - depth, math.ceil(radius)))
hsi, e = np.histogram(siDist, bins=range(math.ceil(radius) - depth, math.ceil(radius)))
ratios = (ho.astype(float)/(ho.astype(float)+hsi.astype(float)))[::-1]
# get ratios slope
depthX  = np.array(range(len(ratios)))
slope,i = np.linalg.lstsq(np.vstack([depthX, np.ones(len(depthX))]).T,ratios)[0]
# compute oxygen ratio
oxygenRatio = slope*depthX + i
## plot and visualize
plt.bar(depthX,ratios, label='distribution')
plt.plot(depthX, oxygenRatio, color='r', label='regression')
plt.legend()
plt.xticks(depthX)
plt.xlabel('depth (A)')
plt.ylabel('Oxygen ratio')
plt.show(block=False)


## get 10 nano-particle of incremental 1 Ang. smaller radius
structures = {0:PDB}
for r in range(1,10):
    # get indexes
    idxs = Geometry.get_satisfactory_records_indexes(PDB.indexes, PDB, 'np.sqrt(x**2 + y**2 + z**2) <=%s'%(radius-r))
    # get pdb copy
    pdb  = PDB.get_copy(indexes=idxs)
    # randomly alter to oxygen 5 angstrom deep
    structures[r] = pdb

## adjust ratio of oxygen of every and each nano-particle
structures[0].export_pdb('multiframe_structure_%i.pdb'%0)
for idx in range(1, len(structures)):
    pdb = structures[idx]
    rad = math.ceil( max(np.sqrt(np.sum(pdb.coordinates**2,1))) )
    for ratio, depthMin in zip(oxygenRatio, depthX):
        if ratio<=0:
            continue
        depthMax = depthMin+1
        indexes = Geometry.get_satisfactory_records_indexes(pdb.indexes, pdb, '(np.sqrt(x**2 + y**2 + z**2)>%s) & (np.sqrt(x**2 + y**2 + z**2)<=%s)'%(rad-depthMax,rad-depthMin))
        oxIdxs  = Information.get_records_indexes_by_attribute_value(indexes, pdb, 'element_symbol', 'O')
        siIdxs  = Information.get_records_indexes_by_attribute_value(indexes, pdb, 'element_symbol', 'Si')
        # find out number of Si or O to transform
        neededOx = int(math.ceil(ratio*len(indexes)))
        neededSi = len(indexes)-neededOx
        if neededOx>len(oxIdxs):
            diff     = neededOx-len(oxIdxs)
            proba    = float(diff)/float(len(siIdxs))
            modIdxs  = [i for i in siIdxs if random.random()<=proba]
            Modify.set_records_attribute_value(modIdxs, pdb, 'element_symbol', 'O')
            Modify.set_records_attribute_value(modIdxs, pdb, 'atom_name', 'O')
        elif neededSi>len(siIdxs):
            diff     = neededSi-len(siIdxs)
            proba    = float(diff)/float(len(oxIdxs))
            modIdxs  = [i for i in oxIdxs if random.random()<=proba]
            Modify.set_records_attribute_value(modIdxs, pdb, 'element_symbol', 'Si')
            Modify.set_records_attribute_value(modIdxs, pdb, 'atom_name', 'Si')
    pdb.export_pdb('multiframe_structure_%i.pdb'%idx)



##########################################################################################
##############################  IMPORTING USEFUL DEFINITIONS  ############################
from __future__ import print_function
import time, multiprocessing
from collections import OrderedDict
import numpy as np


##########################################################################################
#####################################  USER VARIBLES  ####################################
# to get real number of physical cores one should use
# https://pypi.python.org/pypi/cpu_cores instead of multiprocessing.cpu_count()
CORES            = range(1,multiprocessing.cpu_count()+1)# [1,2,3,4]
AVERAGE          = 100
NUMBER_OF_ATOMS  = [1000, 2000, 3000, 4000, 5000, 10000, 20000, 30000, 40000, 50000, 75000, 100000]
HISTOGRAMS       = True
PAIRS_DISTANCES  = True
ATOMIC_DISTANCES = True


##########################################################################################
##################################### INIT VARIABLES #####################################
CORES            = sorted(CORES)
NUMBER_OF_ATOMS  = sorted(NUMBER_OF_ATOMS)
DATA             = {}


##########################################################################################
####################################### HISTOGRAMS #######################################
print('peformance is computed as the mean time of %s runs'%AVERAGE)

if HISTOGRAMS:
    from fullrmc.Core.pairs_histograms import single_pairs_histograms

    print('testing pairs_histograms computations')
    print('=====================================')
    atomIndex = np.int32(100)
    DATA["histograms"] = OrderedDict()
    for SIZE in NUMBER_OF_ATOMS:
        DATA["histograms"][SIZE] = [[],[]]
        distances = 1000*np.random.random((SIZE,)).astype(np.float32)

        moleculeIndex = []
        molIdx = 0
        while len(moleculeIndex)<len(distances):
            add = min( len(distances)/1000, len(distances)-len(moleculeIndex) )
            moleculeIndex.extend( [molIdx]*add )
            molIdx +=1
        moleculeIndex = np.array(moleculeIndex, dtype=np.int32)

        elementIndex = []
        elIdx = 0
        while len(elementIndex)<len(distances):
            add = min( len(distances)/100, len(distances)-len(elementIndex) )
            elementIndex.extend( [elIdx]*add )
            elIdx +=1
        elementIndex = np.array(elementIndex, dtype=np.int32)


        minDistance = np.float32(1)
        maxDistance = np.float32(500)
        bin = np.float32(0.1)
        histSize = np.int32( (maxDistance-minDistance)/bin + 10 )

        for n in CORES:
            mtime = []
            for _ in range(AVERAGE):
                hintra = np.zeros((elIdx,elIdx,histSize), dtype=np.float32)
                hinter = np.zeros((elIdx,elIdx,histSize), dtype=np.float32)
                tic = time.time()
                single_pairs_histograms( atomIndex     = atomIndex,
                                         distances     = distances,
                                         moleculeIndex = moleculeIndex,
                                         elementIndex  = elementIndex,
                                         hintra        = hintra,
                                         hinter        = hinter,
                                         minDistance   = minDistance,
                                         maxDistance   = maxDistance,
                                         bin           = bin,
                                         allAtoms      = True,
                                         ncores        = np.int32(n))
                mtime.append(time.time()-tic)
            print("histograms: %s atoms, ncores %s --> %.10f "%(SIZE, n, sum(mtime)/len(mtime)))
            DATA["histograms"][SIZE][0].append(n)
            DATA["histograms"][SIZE][1].append(sum(mtime)/len(mtime))


##########################################################################################
#################################### PAIRS DISTANCES #####################################
if PAIRS_DISTANCES:
    from fullrmc.Core.pairs_distances import pairs_distances_to_indexcoords, pairs_distances_to_multi_points

    print('\n')
    print('testing pairs_distances computations')
    print('====================================')

    DATA["distance using index"]  = OrderedDict()
    DATA["distance to multiple points"] = OrderedDict()
    for SIZE in NUMBER_OF_ATOMS:
        DATA["distance using index"][SIZE] = [[],[]]
        DATA["distance to multiple points"][SIZE] = [[],[]]
        atomIndex = np.int32(0)
        coords = np.random.random((SIZE, 3)).astype(np.float32)
        points = np.random.random((3,100)).astype(np.float32)

        basis = np.ones((3, 3)).astype(np.float32)

        for n in CORES:
            mtime = []
            for _ in range(AVERAGE):
                tic = time.time()
                pairs_distances_to_indexcoords( atomIndex = atomIndex,
                                                coords    = coords,
                                                basis     = basis,
                                                isPBC     = True,
                                                allAtoms  = True,
                                                ncores    = np.int32(n))
            mtime.append(time.time()-tic)
            print("distance using index: %s atoms, ncores %s --> %.10f "%(SIZE, n, sum(mtime)/len(mtime)))
            DATA["distance using index"][SIZE][0].append(n)
            DATA["distance using index"][SIZE][1].append(sum(mtime)/len(mtime))

        for n in CORES:
            mtime = []
            for _ in range(AVERAGE):
                tic = time.time()
                pairs_distances_to_multi_points( points    = points,
                                                 coords    = coords,
                                                 basis     = basis,
                                                 isPBC     = True,
                                                 ncores    = np.int32(n))
            mtime.append(time.time()-tic)
            print("distance to multiple points: %s atoms, ncores %s --> %.10f "%(SIZE, n, sum(mtime)/len(mtime)))
            DATA["distance to multiple points"][SIZE][0].append(n)
            DATA["distance to multiple points"][SIZE][1].append(sum(mtime)/len(mtime))


##########################################################################################
#################################### ATOMIC DISTANCES ####################################
if ATOMIC_DISTANCES:
    from fullrmc.Core.atomic_distances import single_atomic_distances_dists

    print('\n')
    print('testing atomic_distances computations')
    print('=====================================')

    atomIndex = np.int32(0)
    distances = 1000*np.random.random((SIZE, )).astype(np.float32)


    DATA["intermolecular distances"] = OrderedDict()
    for SIZE in NUMBER_OF_ATOMS:
        DATA["intermolecular distances"][SIZE] = [[],[]]

        moleculeIndex = []
        molIdx = 0
        while len(moleculeIndex)<len(distances):
            add = min( len(distances)/1000, len(distances)-len(moleculeIndex) )
            moleculeIndex.extend( [molIdx]*add )
            molIdx +=1
        moleculeIndex = np.array(moleculeIndex, dtype=np.int32)

        elementIndex = []
        elIdx = 0
        while len(elementIndex)<len(distances):
            add = min( len(distances)/100, len(distances)-len(elementIndex) )
            elementIndex.extend( [elIdx]*add )
            elIdx +=1
        elementIndex = np.array(elementIndex, dtype=np.int32)

        lowerLimit = np.ones((elIdx,elIdx,1), dtype=np.float32)
        upperLimit = 3*np.ones((elIdx,elIdx,1), dtype=np.float32)

        for n in CORES:
            mtime = []
            for _ in range(AVERAGE):
                dintra = np.zeros((elIdx,elIdx,1), dtype=np.float32)
                dinter = np.zeros((elIdx,elIdx,1), dtype=np.float32)
                nintra = np.zeros((elIdx,elIdx,1), dtype=np.int32)
                ninter = np.zeros((elIdx,elIdx,1), dtype=np.int32)

                tic = time.time()
                single_atomic_distances_dists( atomIndex             = atomIndex,
                                               distances             = distances,
                                               moleculeIndex         = moleculeIndex,
                                               elementIndex          = elementIndex,
                                               dintra                = dintra,
                                               dinter                = dinter,
                                               nintra                = nintra,
                                               ninter                = ninter,
                                               lowerLimit            = lowerLimit,
                                               upperLimit            = upperLimit,
                                               interMolecular        = True,
                                               intraMolecular        = True,
                                               countWithinLimits     = True,
                                               reduceDistanceToUpper = False,
                                               reduceDistanceToLower = False,
                                               reduceDistance        = False,
                                               allAtoms              = True,
                                               ncores                = np.int32(n) )
            mtime.append(time.time()-tic)
            print("intermolecular distances: %s atoms, ncores %s --> %.10f "%(SIZE, n, sum(mtime)/len(mtime)))
            DATA["intermolecular distances"][SIZE][0].append(n)
            DATA["intermolecular distances"][SIZE][1].append(sum(mtime)/len(mtime))



##########################################################################################
########################################## PLOT ##########################################
import matplotlib.pyplot as plt
colormap = plt.cm.jet
colors = [colormap(i) for i in np.linspace(0, 1,len(NUMBER_OF_ATOMS))]

for kdict in DATA:
    vdict = DATA[kdict]
    fig = plt.figure()
    fig.patch.set_facecolor('white')
    fig.suptitle(kdict)
    maxs = []
    mins = []
    for i, k in enumerate(vdict):
        v = vdict[k]
        maxs.append(max(v[1]))
        mins.append(min(v[1]))
        plt.plot(v[0], v[1], marker='o', color=colors[i], label="%s atoms"%k)
    plt.gca().set_xlabel("number of cores")
    plt.gca().set_ylabel("time (s)")
    plt.gca().legend(frameon=False, ncol=np.ceil(len(NUMBER_OF_ATOMS)/6.).astype(int) )
    plt.gca().set_xlim([0,CORES[-1]+1])
    mins = min(mins)
    maxs = max(maxs)
    plt.gca().set_ylim([mins-0.1*mins, maxs+0.1*maxs])
plt.show()
exit()



"""
This is a C compiled module to compute atomic bonds.
"""
import cython
cimport cython
import numpy as np
cimport numpy as np
from numpy cimport ndarray
from cython.parallel import prange
from fullrmc.Core.pairs_distances import pairs_distances_to_point

# declare types
NUMPY_FLOAT32 = np.float32
NUMPY_INT32   = np.int32
ctypedef np.float32_t C_FLOAT32
ctypedef np.int32_t   C_INT32

# declare constants
cdef C_FLOAT32 FLOAT_ZERO  = 0.0
cdef C_FLOAT32 FLOAT_ONE   = 1.0
cdef C_INT32   INT32_ZERO  = 0
cdef C_INT32   INT32_ONE   = 1



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
cdef C_FLOAT32 _single_atom_single_shell_dists( C_FLOAT32[:] distances,
                                                C_FLOAT32    lowerShell,
                                                C_FLOAT32    upperShell,
                                                C_INT32      ncores = 1) nogil:
    # declare variables
    cdef C_INT32 i
    cdef C_FLOAT32 coordNumber = FLOAT_ZERO
    cdef C_INT32 num_threads = ncores
    # loop
    for i in prange(INT32_ZERO, <C_INT32>distances.shape[0], INT32_ONE, nogil=True, schedule="static", num_threads=num_threads):
    #for i from 0 <= i < <C_INT32>distances.shape[0]:
        if lowerShell <= distances[i] <= upperShell:
            coordNumber += FLOAT_ONE
            #with gil:
            #    print(i, lowerShell, upperShell, distances[i])
    return coordNumber



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_single_shell_subdists( C_FLOAT32[:] distances,
                                       C_FLOAT32    lowerShell,
                                       C_FLOAT32    upperShell,
                                       C_INT32      ncores = 1):
    return _single_atom_single_shell_dists( distances  = distances,
                                            lowerShell = lowerShell,
                                            upperShell = upperShell,
                                            ncores     = ncores)



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_single_shell_totdists( ndarray[C_FLOAT32, ndim=1] distances,
                                       C_INT32[:]                 shellIndexes,
                                       C_FLOAT32                  lowerShell,
                                       C_FLOAT32                  upperShell,
                                       C_INT32                    ncores = 1):
    # declare variables
    return _single_atom_single_shell_dists( distances  = distances[shellIndexes],
                                            lowerShell = lowerShell,
                                            upperShell = upperShell,
                                            ncores     =  ncores)



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_single_shell_coords( C_INT32                    coreIndex,
                                     C_INT32[:]                 shellIndexes,
                                     ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                     C_FLOAT32[:,:]             basis not None,
                                     bint                       isPBC,
                                     C_FLOAT32                  lowerShell,
                                     C_FLOAT32                  upperShell,
                                     C_INT32                    ncores = 1):
    # declare variables
    distances = pairs_distances_to_point( point  = boxCoords[ coreIndex ],
                                          coords = boxCoords[ shellIndexes ],
                                          basis  = basis,
                                          isPBC  = isPBC,
                                          ncores = ncores)
    # compute and return
    return single_atom_single_shell_subdists(distances  = distances,
                                             lowerShell = lowerShell,
                                             upperShell = upperShell,
                                             ncores     = ncores)



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_multi_shells_totdists( ndarray[C_FLOAT32, ndim=1] distances,
                                       list                       shellsIndexes,
                                       C_FLOAT32[:]               lowerShells,
                                       C_FLOAT32[:]               upperShells,
                                       C_INT32                    ncores = 1):
    # declare variables
    cdef C_INT32 i
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] coordNumbers = np.zeros(len(shellsIndexes), dtype=NUMPY_FLOAT32)
    # compute coordination numbers
    for i from 0 <= i < <C_INT32>len(shellsIndexes):
        coordNumbers[i] = _single_atom_single_shell_dists( distances   = distances[shellsIndexes[i]],
                                                           lowerShell  = lowerShells[i],
                                                           upperShell  = upperShells[i],
                                                           ncores      = ncores)
    return coordNumbers



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_multi_shells_coords( C_INT32                    coreIndex,
                                     list                       shellsIndexes,
                                     ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                     C_FLOAT32[:,:]             basis not None,
                                     bint                       isPBC,
                                     C_FLOAT32[:]               lowerShells,
                                     C_FLOAT32[:]               upperShells,
                                     C_INT32                    ncores = 1):
    # declare variables
    cdef C_INT32 i
    cdef ndarray[C_FLOAT32,  mode="c", ndim=1] coordNumbers = np.zeros(len(shellsIndexes), dtype=NUMPY_FLOAT32)
    # compute coordination numbers
    for i from 0 <= i < <C_INT32>len(shellsIndexes):
        # declare variables
        distances = pairs_distances_to_point( point  = boxCoords[ coreIndex ],
                                              coords = boxCoords[ shellsIndexes[i] ],
                                              basis  = basis,
                                              isPBC  = isPBC,
                                              ncores = ncores)
        coordNumbers[i] = _single_atom_single_shell_dists( distances   = distances,
                                                           lowerShell  = lowerShells[i],
                                                           upperShell  = upperShells[i],
                                                           ncores      = ncores)
    return coordNumbers



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_coord_number_totdists( C_INT32                    atomIndex,
                                       ndarray[C_FLOAT32, ndim=1] distances,
                                       list                       coresIndexes,
                                       list                       shellsIndexes,
                                       list                       lowerShells,
                                       list                       upperShells,
                                       list                       asCoreDefIdxs,
                                       list                       inShellDefIdxs,
                                       C_FLOAT32[:]               coordNumData,
                                       C_INT32                    ncores = 1):
    # declare variables
    cdef C_INT32 defIdx
    # compute coordination numbers as core
    for defIdx in asCoreDefIdxs[atomIndex]:
        coordNumber = single_atom_single_shell_totdists(distances    = distances,
                                                        shellIndexes = shellsIndexes[defIdx],
                                                        lowerShell   = lowerShells[defIdx] ,
                                                        upperShell   = upperShells[defIdx],
                                                        ncores       = ncores)
        coordNumData[defIdx] += coordNumber
    # compute coordination numbers in shell
    for defIdx in asCoreDefIdxs[atomIndex]:
        coordNumber = single_atom_single_shell_totdists(distances    = distances,
                                                        shellIndexes = coresIndexes[defIdx],
                                                        lowerShell   = lowerShells[defIdx] ,
                                                        upperShell   = upperShells[defIdx],
                                                        ncores       = ncores)
        coordNumData[defIdx] += coordNumber



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def single_atom_coord_number_coords( C_INT32                    atomIndex,
                                     ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                     C_FLOAT32[:,:]             basis not None,
                                     bint                       isPBC,
                                     list                       coresIndexes,
                                     list                       shellsIndexes,
                                     list                       lowerShells,
                                     list                       upperShells,
                                     list                       asCoreDefIdxs,
                                     list                       inShellDefIdxs,
                                     C_FLOAT32[:]               coordNumData,
                                     C_INT32                    ncores = 1):
    # compute coordination numbers as core
    for defIdx in asCoreDefIdxs[atomIndex]:
        coordNumber = single_atom_single_shell_coords( coreIndex    = atomIndex,
                                                       shellIndexes = shellsIndexes[defIdx],
                                                       boxCoords    = boxCoords,
                                                       basis        = basis,
                                                       isPBC        = isPBC,
                                                       lowerShell   = lowerShells[defIdx],
                                                       upperShell   = upperShells[defIdx],
                                                       ncores       = ncores)
        coordNumData[defIdx] += coordNumber
    # compute coordination numbers in shell
    for defIdx in inShellDefIdxs[atomIndex]:
        coordNumber = single_atom_single_shell_coords( coreIndex    = atomIndex,
                                                       shellIndexes = coresIndexes[defIdx],
                                                       boxCoords    = boxCoords,
                                                       basis        = basis,
                                                       isPBC        = isPBC,
                                                       lowerShell   = lowerShells[defIdx],
                                                       upperShell   = upperShells[defIdx],
                                                       ncores       = ncores)
        coordNumData[defIdx] += coordNumber



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def multi_atoms_coord_number_totdists( C_INT32[:]   indexes,
                                       list         distances,
                                       list         coresIndexes,
                                       list         shellsIndexes,
                                       list         lowerShells,
                                       list         upperShells,
                                       list         asCoreDefIdxs,
                                       list         inShellDefIdxs,
                                       C_FLOAT32[:] coordNumData,
                                       C_INT32      ncores = 1):
    # declare variables
    cdef C_INT32 i
    for i from 0 <= i < <C_INT32>len(indexes):
        single_atom_coord_number_totdists( atomIndex      = indexes[i],
                                           distances      = distances[i],
                                           coresIndexes   = coresIndexes,
                                           shellsIndexes  = shellsIndexes,
                                           lowerShells    = lowerShells,
                                           upperShells    = upperShells,
                                           asCoreDefIdxs  = asCoreDefIdxs,
                                           inShellDefIdxs = inShellDefIdxs,
                                           coordNumData   = coordNumData,
                                           ncores         = ncores)



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def multi_atoms_coord_number_coords( C_INT32[:]                 indexes,
                                     ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                     C_FLOAT32[:,:]             basis not None,
                                     bint                       isPBC,
                                     list                       coresIndexes,
                                     list                       shellsIndexes,
                                     list                       lowerShells,
                                     list                       upperShells,
                                     list                       asCoreDefIdxs,
                                     list                       inShellDefIdxs,
                                     C_FLOAT32[:]               coordNumData,
                                     C_INT32                    ncores = 1):
    # declare variables
    cdef C_INT32 i
    for i from 0 <= i < <C_INT32>len(indexes):
        single_atom_coord_number_coords( atomIndex      = indexes[i],
                                         boxCoords      = boxCoords,
                                         basis          = basis,
                                         isPBC          = isPBC,
                                         coresIndexes   = coresIndexes,
                                         shellsIndexes  = shellsIndexes,
                                         lowerShells    = lowerShells,
                                         upperShells    = upperShells,
                                         asCoreDefIdxs  = asCoreDefIdxs,
                                         inShellDefIdxs = inShellDefIdxs,
                                         coordNumData   = coordNumData,
                                         ncores         = ncores)





@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def all_atoms_coord_number_totdists( ndarray[C_FLOAT32, ndim=2] distances,
                                     list                       coresIndexes,
                                     list                       shellsIndexes,
                                     list                       lowerShells,
                                     list                       upperShells,
                                     list                       asCoreDefIdxs,
                                     list                       inShellDefIdxs,
                                     C_FLOAT32[:]               coordNumData,
                                     C_INT32                    ncores = 1):
    # declare variables
    cdef C_INT32 i
    # get indexes
    cdef ndarray[C_INT32,  mode="c", ndim=1] indexes = np.arange(<C_INT32>distances.shape[1], dtype=NUMPY_INT32)
    # run multiple atoms coordination number using distances
    multi_atoms_coord_number_totdists( indexes        = indexes,
                                       distances      = distances,
                                       coresIndexes   = coresIndexes,
                                       shellsIndexes  = shellsIndexes,
                                       lowerShells    = lowerShells,
                                       upperShells    = upperShells,
                                       asCoreDefIdxs  = asCoreDefIdxs,
                                       inShellDefIdxs = inShellDefIdxs,
                                       coordNumData   = coordNumData,
                                       ncores         = ncores)



@cython.nonecheck(False)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.always_allow_keywords(False)
def all_atoms_coord_number_coords( ndarray[C_FLOAT32, ndim=2] boxCoords not None,
                                   C_FLOAT32[:,:]             basis not None,
                                   bint                       isPBC,
                                   list                       coresIndexes,
                                   list                       shellsIndexes,
                                   list                       lowerShells,
                                   list                       upperShells,
                                   list                       asCoreDefIdxs,
                                   list                       inShellDefIdxs,
                                   C_FLOAT32[:]               coordNumData,
                                   C_INT32                    ncores = 1):
    # declare variables
    cdef C_INT32 i
    # get indexes
    cdef ndarray[C_INT32,  mode="c", ndim=1] indexes = np.arange(<C_INT32>boxCoords.shape[0], dtype=NUMPY_INT32)
    # run multiple atoms coordination number using coordinates
    multi_atoms_coord_number_coords( indexes        = indexes,
                                     boxCoords      = boxCoords,
                                     basis          = basis,
                                     isPBC          = isPBC,
                                     coresIndexes   = coresIndexes,
                                     shellsIndexes  = shellsIndexes,
                                     lowerShells    = lowerShells,
                                     upperShells    = upperShells,
                                     asCoreDefIdxs  = asCoreDefIdxs,
                                     inShellDefIdxs = inShellDefIdxs,
                                     coordNumData   = coordNumData,
                                     ncores         = ncores)



"""
Agitations contains all MoveGenerator classes that agitate and shake
structures such as distances, angles, etc.

.. inheritance-diagram:: fullrmc.Generators.Agitations
    :parts: 1

+------------------------------------------------------+------------------------------------------------------+
|.. figure:: distanceAgitation.png                     |.. figure:: angleAgitation.png                        |
|   :width: 375px                                      |   :width: 375px                                      |
|   :height: 300px                                     |   :height: 300px                                     |
|   :align: left                                       |   :align: left                                       |
|                                                      |                                                      |
|   Random H-H bond length agitations generated on     |   Random H-O-H angle agitation generated on water    |
|   dihydrogen molecules. At room temperature, H2      |   molecules. At room temperature, water molecule     |
|   molecule bond length fluctuates around 0.74        |   angle formed between the two vectors formed between|
|   Angstroms. Red hydrogen atoms represent the shrank |   consecutively the Oxygen atom and the two hydrogen |
|   H-H bond length molecule while blue hydrogen atoms |   atoms is about 105 deg. Shrank H-O-H angles are    |
|   represent the expanded H-H bond length molecules.  |   represented by the red hydrogen while expanded     |
|                                                      |   angles are represented in blue.                    |
|   (:class:`DistanceAgitationGenerator`)              |   (:class:`AngleAgitationGenerator`)                 |
+------------------------------------------------------+------------------------------------------------------+

.. raw:: html

    <iframe width="560" height="315"
    src="https://www.youtube.com/embed/qTJux9kZCOo?rel=0"
    frameborder="0" allowfullscreen>
    </iframe>

"""
# standard libraries imports
from __future__ import print_function

# external libraries imports
import numpy as np

# fullrmc imports
from ..Globals import INT_TYPE, FLOAT_TYPE, PI, LOGGER
from ..Globals import str, long, unicode, bytes, basestring, range, xrange, maxint
from ..Core.Collection import is_number, is_integer, get_rotation_matrix, generate_random_float, generate_random_vector
from ..Core.MoveGenerator import MoveGenerator, PathGenerator


class DistanceAgitationGenerator(MoveGenerator):
    """
    Generates random agitation moves upon a distance separating two atoms
    by translating both atoms away from each other or closer to each other
    along the direction line between them. This is mainly used to shake two
    atoms bond distance by increasing and decreasing the bond length.
    Only groups of length 2 are accepted.

    :Parameters:
        #. group (None, Group): The group instance. It must contain exactly
           two indexes.
        #. amplitude (number):  The maximum translation amplitude in Angstroms
           applied on every atom.
        #. symmetric (bool): Whether to apply the same amplitude of
           translation on both atoms or not.
        #. shrink (None, bool): Whether to always shrink the distance or
           expand it.
           If True, moves will always bring atoms closer to each other.
           If False, moves will always bring atoms away from each other.
           If None, no orientation is forced, therefore atoms can randomly get
           closer to each other or away from each other.
        #. agitate (tuple): It's a tuple of two boolean values, at least one
           of them must be True.
           Whether to agitate the first atom, the second or both. This is
           useful to set an atom fixed while only
           the other succumb the agitation to adjust the distance. For
           instance in a C-H group it can be useful and
           logical to adjust the bond length by moving only the hydrogen atom
           along the bond direction.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Agitations import DistanceAgitationGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to random agitations of distance seperating two atoms.
        # Maximum agitation amplitude is set to to 0.5A
        for g in ENGINE.groups:
            if len(g)==2:
                g.set_move_generator( DistanceAgitationGenerator(amplitude=0.5) )

    """
    def __init__(self, group=None, amplitude=0.2, symmetric=True, shrink=None, agitate=(True,True)):
        super(DistanceAgitationGenerator, self).__init__(group=group)
        # set amplitude
        self.set_amplitude(amplitude)
        # set symmetric
        self.set_symmetric(symmetric)
        # set shrink
        self.set_shrink(shrink)
        # set agitated
        self.set_agitate(agitate)

    @property
    def amplitude(self):
        """Maximum agitation amplitude."""
        return self.__amplitude

    @property
    def shrink(self):
        """Shrink flag value."""
        return self.__shrink

    @property
    def symmetric(self):
        """Symmetric flag value."""
        return self.__symmetric

    @property
    def agitate(self):
        """Agitate tuple flags value."""
        return self.__agitate

    def check_group(self, group):
        """
        Check the generator's group.

        :Parameters:
            #. group (Group): the Group instance.
        """
        if len(group.indexes)!=2:
            return False, "two atoms are needed in a group to perform distance agitation movements."
        else:
            return True, ""

    def set_amplitude(self, amplitude):
        """
        Sets maximum translation vector allowed amplitude.

        :Parameters:
            #. amplitude (number): the maximum allowed translation
               vector amplitude.
        """
        assert is_number(amplitude), LOGGER.error("Translation amplitude must be a number")
        amplitude = float(amplitude)
        assert amplitude>0, LOGGER.error("Translation amplitude must be bigger than 0")
        self.__amplitude = FLOAT_TYPE(amplitude)

    def set_symmetric(self, symmetric):
        """
        Set symmetric flag value.

        :Parameters:
            #. symmetric (bool): Whether to apply the same amplitude of
               translation on both atoms or not.
        """
        assert isinstance(symmetric, bool), LOGGER.error("symmetric must be boolean")
        self.__symmetric = symmetric

    def set_shrink(self, shrink):
        """
        Set shrink flag value.

        :Parameters:
            #. shrink (None, bool): Whether to always shrink the distance
               or expand it.
               If True, moves will always bring atoms closer to each other.
               If False, moves will always bring atoms away from each other.
               If None, no orientation is forced, therefore distance can
               increase or decrease randomly at every step.
        """
        assert shrink in (None, True, False), LOGGER.error("shrink can only be None, True or False")
        self.__shrink = shrink

    def set_agitate(self, agitate):
        """
        Set agitate tuple value.

        :Parameters:
            #. agitate (tuple): It's a tuple of two boolean values, at least
               one of them must be True. Whether to agitate the first atom,
               the second or both. This is useful to set an atom fixed while
               only the other succumb the agitation to adjust the distance.
               For instance in a C-H group it can be useful and logical to
               adjust the bond length by moving only the hydrogen atom along
               the bond direction.
        """
        assert isinstance(agitate, (list,tuple)), LOGGER.error("agitate must be a list or a tuple")
        assert len(agitate)==2, LOGGER.error("agitate must have 2 items")
        assert [isinstance(a,bool) for a in agitate]==[True,True], LOGGER.error("agitate items must be boolean")
        assert agitate[0] or agitate[1], LOGGER.error("agitate both items can't be False")
        self.__agitate = (agitate[0], agitate[1])

    def transform_coordinates(self, coordinates, argument=None):
        """
        Translate coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the translation.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the translation.
            #. argument (object): Any python object. Not used in this
               generator.
        """
        if coordinates.shape[0]!=2:
            # atoms where removed, fall back to random translation
            return coordinates+generate_random_vector(minAmp=self.__amplitude[0],
                                                      maxAmp=self.__amplitude[1])
        else:
            # get normalized direction vector
            vector  = FLOAT_TYPE( coordinates[0,:]-coordinates[1,:] )
            vector /= FLOAT_TYPE( np.linalg.norm(vector) )
            # create amplitudes
            if self.__symmetric:
                amp0 = amp1 =  FLOAT_TYPE(generate_random_float()*self.__amplitude)
            else:
                amp0 =  FLOAT_TYPE(generate_random_float()*self.__amplitude)
                amp1 =  FLOAT_TYPE(generate_random_float()*self.__amplitude)
            # create shrink flag
            if self.__shrink is None:
                shrink = (1-2*generate_random_float())>0
            else:
                shrink = self.__shrink
            # create directions
            if shrink:
                dir0 = FLOAT_TYPE(-1)
                dir1 = FLOAT_TYPE( 1)
            else:
                dir0 = FLOAT_TYPE( 1)
                dir1 = FLOAT_TYPE(-1)
            # create translation vectors
            translationVectors      = np.empty((2,3), dtype=FLOAT_TYPE)
            translationVectors[0,:] = self.__agitate[0]*dir0*amp0*vector
            translationVectors[1,:] = self.__agitate[1]*dir1*amp1*vector
            # translate and return
            return coordinates+translationVectors


class AngleAgitationGenerator(MoveGenerator):
    """
    Generate random agitation moves upon an angle defined between two vectors
    left-central and right-central where (central, left, right) are three
    atoms. Move will be performed on left and/or right atom while central
    atom will always remain fixed. Distances between left/right and central
    atoms will remain unchanged. This is mainly used to shake bonded atoms
    angles by increasing and decreasing the bond length.
    Only groups of length 3 are accepted.

    :Parameters:
        #. group (None, Group): The group instance. It must contain exactly
           three indexes in respective order (central, left, right)
           atoms index.
        #. amplitude (number):  The maximum agitation angle amplitude in
           degrees of left and right atoms separately.
        #. symmetric (bool): Whether to apply the same amplitude of rotation
           on both left and right atoms or not.
        #. shrink (None, bool): Whether to always shrink the angle or expand it.
           If True, moves will always reduce angle.
           If False, moves will always increase angle.
           If None, no orientation is forced, therefore angle can randomly
           get wider or tighter.
        #. agitate (tuple): It's a tuple of two boolean values for
           respectively (left, right) atoms,
           at least one of them must be True. Whether to agitate the left
           atom, the right or both. This is useful to set an atom fixed while
           only the other succumb the agitation to adjust the angle.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Generators.Agitations import AngleAgitationGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...
        # Re-define groups selector if needed ...

        # set moves generators to random agitations of the angle formed between
        # one central atom and other two. Maximum agitation amplitude is set to to 10.
        for g in ENGINE.groups:
            if len(g)==3:
                g.set_move_generator( AngleAgitationGenerator(amplitude=10) )

    """
    def __init__(self, group=None, amplitude=2, symmetric=True, shrink=None, agitate=(True,True)):
        super(AngleAgitationGenerator, self).__init__(group=group)
        # set amplitude
        self.set_amplitude(amplitude)
        # set symmetric
        self.set_symmetric(symmetric)
        # set shrink
        self.set_shrink(shrink)
        # set agitated
        self.set_agitate(agitate)

    @property
    def amplitude(self):
        """Maximum agitation angle amplitude in rad."""
        return self.__amplitude

    @property
    def shrink(self):
        """Shrink flag value."""
        return self.__shrink

    @property
    def symmetric(self):
        """Symmetric flag value."""
        return self.__symmetric

    @property
    def agitate(self):
        """Agitate tuple flags value."""
        return self.__agitate

    def check_group(self, group):
        """
        Check the generator's group.

        :Parameters:
            #. group (Group): the Group instance.
        """
        if len(group.indexes)!=3:
            return False, "Exactly three atoms are needed in a group to perform angle agitation movements."
        else:
            return True, ""

    def set_amplitude(self, amplitude):
        """
        Set maximum allowed agitation rotation angle amplitude in degrees of
        left and right atoms separately and transforms it to rad.

        :Parameters:
            #. amplitude (number):  The maximum agitation angle amplitude in
               degrees of left and right atoms separately.
        """
        assert is_number(amplitude), LOGGER.error("Agitation angle amplitude must be a number")
        amplitude = float(amplitude)
        assert amplitude>0, LOGGER.error("Agitation angle amplitude must be bigger than 0")
        assert amplitude<=90, LOGGER.error("Agitation angle amplitude must be smaller than 90")
        self.__amplitude = FLOAT_TYPE(amplitude*PI/180.)

    def set_symmetric(self, symmetric):
        """
        Set symmetric flag value.

        :Parameters:
            #. symmetric (bool): Whether to apply the same amplitude of
               translation on both atoms or not.
        """
        assert isinstance(symmetric, bool), LOGGER.error("symmetric must be boolean")
        self.__symmetric = symmetric

    def set_shrink(self, shrink):
        """
        Set shrink flag value.

        :Parameters:
            #. shrink (None, bool): Whether to always shrink the distance or
               expand it.
               If True, moves will always bring atoms closer to each other.
               If False, moves will always bring atoms away from each other.
               If None, no orientation is forced, therefore distance can
               increase or decrease randomly at every step.
        """
        assert shrink in (None, True, False), LOGGER.error("shrink can only be None, True or False")
        self.__shrink = shrink

    def set_agitate(self, agitate):
        """
        Set agitate tuple value.

        :Parameters:
            #. agitate (tuple): It's a tuple of two boolean values, at least
               one of them must be True. Whether to agitate the first atom,
               the second or both. This is useful to set an atom fixed while
               only the other succumb the agitation to adjust the distance.
               For instance in a C-H group it can be useful and logical to
               adjust the bond length by moving only the hydrogen atom along
               the bond direction.
        """
        assert isinstance(agitate, (list,tuple)), LOGGER.error("agitate must be a list or a tuple")
        assert len(agitate)==2, LOGGER.error("agitate must have 2 items")
        assert [isinstance(a,bool) for a in agitate]==[True,True], LOGGER.error("agitate items must be boolean")
        assert agitate[0] or agitate[1], LOGGER.error("agitate both items can't be False")
        self.__agitate = (agitate[0], agitate[1])

    def transform_coordinates(self, coordinates, argument=None):
        """
        Translate coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the translation.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the translation.
            #. argument (object): Any python object. Not used in this generator.
        """
        if coordinates.shape[0]!=3:
            # atoms where removed, fall back to random translation
            return coordinates+generate_random_vector(minAmp=self.__amplitude[0],
                                                      maxAmp=self.__amplitude[1])
        else:
            # get atoms group center
            center = np.sum(coordinates, 0)/coordinates.shape[0]
            # translate to origin
            rotatedCoordinates = coordinates-center
            # get normalized direction vectors
            leftVector   = FLOAT_TYPE( rotatedCoordinates[1,:]-rotatedCoordinates[0,:] )
            leftVector  /= FLOAT_TYPE( np.linalg.norm(leftVector) )
            rightVector  = FLOAT_TYPE( rotatedCoordinates[2,:]-rotatedCoordinates[0,:] )
            rightVector /= FLOAT_TYPE( np.linalg.norm(rightVector) )
            # get rotation axis
            rotationAxis = np.cross(leftVector, rightVector)
            if rotationAxis[0]==rotationAxis[1]==rotationAxis[2]==0.:
                rotationAxis = np.array(1-2*np.random.random(3), dtype=FLOAT_TYPE)
                rotationAxis /= FLOAT_TYPE( np.linalg.norm(rotationAxis) )
            # create shrink flag
            if self.__shrink is None:
                shrink = (1-2*generate_random_float())>0
            else:
                shrink = self.__shrink
            # get rotation angles
            if self.__symmetric:
                angleLeft  = angleRight = FLOAT_TYPE(generate_random_float()*self.__amplitude)
            else:
                angleLeft  = FLOAT_TYPE(generate_random_float()*self.__amplitude)
                angleRight = FLOAT_TYPE(generate_random_float()*self.__amplitude)
            # create directions
            if shrink:
                angleLeft  *= FLOAT_TYPE(-1)
                angleRight *= FLOAT_TYPE( 1)
            else:
                angleLeft  *= FLOAT_TYPE( 1)
                angleRight *= FLOAT_TYPE(-1)
            # rotate
            if self.__agitate[0]:
                rotationMatrix = get_rotation_matrix(rotationAxis, angleLeft)
                rotatedCoordinates[1,:] = np.dot( rotationMatrix, rotatedCoordinates[1,:])
            if self.__agitate[1]:
                rotationMatrix = get_rotation_matrix(rotationAxis, angleRight)
                rotatedCoordinates[2,:] = np.dot( rotationMatrix, rotatedCoordinates[2,:])
            # translate back from center and return
            return np.array(rotatedCoordinates+center, dtype=FLOAT_TYPE)



"""
MoveGenerator contains parent classes for all move generators.
A MoveGenerator sub-class is used at fullrmc's stochastic engine runtime to
generate moves upon selected groups.
Every group has its own MoveGenerator class and definitions, therefore it is
possible to fully customize how a group of atoms should move.

.. inheritance-diagram:: fullrmc.Core.MoveGenerator
    :parts: 1
"""
# standard libraries imports
from __future__ import print_function
import inspect
from random import randint, shuffle

# external libraries imports
import numpy as np

# fullrmc imports
from ..Globals import INT_TYPE, FLOAT_TYPE, LOGGER
from ..Globals import str, long, unicode, bytes, basestring, range, xrange, maxint
from ..Core.Collection import ListenerBase, is_number, is_integer, get_path, generate_random_float
from ..Core.Collection import _Container


#class MoveGenerator(ListenerBase):
class MoveGenerator(object):
    """
    It is the parent class of all moves generators.
    This class can't be instantiated but its sub-classes might be.

    :Parameters:
        #. group (None, Group): The group instance.
    """
    def __init__(self, group=None):
        # init ListenerBase
        super(MoveGenerator, self).__init__()
        # set group
        self.set_group(group)

    @property
    def group(self):
        """ Group instance."""
        return self.__group

    def set_group(self, group):
        """
        Set the MoveGenerator group.

        :Parameters:
            #. group (None, Group): Group instance.
        """
        if group is not None:
            from fullrmc.Core.Group import Group
            assert isinstance(group, Group), LOGGER.error("group must be a fullrmc Group instance")
            valid, message = self.check_group(group)
            if not valid:
                raise Exception( LOGGER.error("%s"%message) )
        self.__group = group

    def check_group(self, group):
        """
        Check the generator's group.
        This method must be overloaded in all MoveGenerator sub-classes.

        :Parameters:
            #. group (Group): the Group instance
        """
        raise Exception(LOGGER.impl("MovesGenerator '%s' method must be overloaded"%inspect.stack()[0][3]))

    def transform_coordinates(self, coordinates, argument=None):
        """
        Transform coordinates. This method is called to move atoms.
        This method must be overloaded in all MoveGenerator sub-classes.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the move.
            #. argument (object): Any other argument needed to perform the
               move. In General it's not needed.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the move.
        """
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def move(self, coordinates):
        """
        Moves coordinates.
        This method must NOT be overloaded in MoveGenerator sub-classes.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the transformation.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the transformation.
        """
        return self.transform_coordinates(coordinates=coordinates)


class RemoveGenerator(MoveGenerator):
    """
    This is a very particular move generator that will not generate moves on
    atoms but removes them from the atomic configuration using a general
    collector mechanism. Remove generators must be used to create defects in
    the simulated system. When the standard error is high, removing atoms
    might reduce the total fit standard error but this can be illusional and
    very limiting because artificial non physical voids can get created in
    the system which will lead to an impossibility to finding a solution at
    the end. It's strongly recommended to exhaust all ideas and possibilities
    in finding a good solution prior to start removing atoms unless structural
    defects is the goal of the simulation.\n
    All removed or amputated atoms are collected by the engine and will
    become available to be re-inserted in the system if needed. But keep in
    mind, it might be physically easy to remove and atom but an impossibility
    to add it back especially if the created voids are smeared out.\n
    Removers are called generators but they behave like selectors. Instead
    of applying a certain move on a group of atoms, they normally pick atoms
    from defined atoms list and apply no moves on those. 'move' and
    'transform_coordinates' methods are not implemented in this class of
    generators and a usage error will be raised if called. 'pick_from_list'
    method is used instead and must be overloaded by all RemoveGenerator
    subclasses.

    **N.B. This class can't be instantiated but its sub-classes might be.**

    :Parameters:
        #. group (None, Group): The group instance which is this case must be
           fullrmc EmptyGroup.
        #. maximumCollected (None, Integer): The maximum number allowed of
           atoms to be removed and collected from atomic configuration by the
           stochastic engine. This property is general to the system and
           checks engine's collected atoms not the number of removed atoms
           via this generator. If None is given, the remover will not
           check for the number of already removed atoms before attempting
           a remove.
        #. allowFittingScaleFactor (bool): Constraints and especially
           experimental ones have a scale factor constant that can be fit.
           Fitting a scale factor happens at stochastic engine's runtime
           at a certain fitting frequency. If this flag set to True,
           then fitting the scale factor will be allowed upon removing atoms.
           When set to False, fitting the constraint scale factor will be
           forbidden upon removing atoms. By default, allowFittingScaleFactor
           is set to False because it's more logical to allow removing only
           atoms that enhances the total standard error without rescaling
           the model's data.
        #. atomsList (None,list,set,tuple,np.ndarray): The list of atomss
           index to chose and remove from.
    """
    def __init__(self, group=None, maximumCollected=None, allowFittingScaleFactor=False, atomsList=None):
        if self.__class__.__name__ == "RemoveGenerator":
            raise Exception(LOGGER.usage("%s instanciation is not allowed"%(self.__class__.__name__)))
        super(RemoveGenerator, self).__init__(group=group)
        # set collectorState
        self._collectorState = None
        # set maximum collected
        self.set_maximum_collected(maximumCollected)
        # set maximum collected
        self.set_allow_fitting_scale_factor(allowFittingScaleFactor)
        # set maximum collected
        self.set_atoms_list(atomsList)

    @property
    def atomsList(self):
        """Atoms list from which atoms will be picked to attempt removal."""
        return self.__atomsList

    @property
    def allowFittingScaleFactor(self):
        """Whether to allow constraints to fit their scale factor upon
        removing atoms."""
        return self.__allowFittingScaleFactor

    @property
    def maximumCollected(self):
        """Maximum collected atoms allowed."""
        return self.__maximumCollected

    def check_group(self, group):
        """
        Check the generator's group.

        :Parameters:
            #. group (Group): The group instance.
        """
        from fullrmc.Core.Group import EmptyGroup
        if isinstance(group, EmptyGroup):
            return True, ""
        else:
            return False, "Only fullrmc EmptyGroup is allowed for CollectorGenerator"

    def set_maximum_collected(self, maximumCollected):
        """
        Set maximum collected number of atoms allowed.

        :Parameters:
            #. maximumCollected (None, Integer): The maximum number allowed of
               atoms to be removed and collected from atomic configuration by
               the stochastic engine. This property is general to the system and
               checks engine's collected atoms not the number of removed atoms
               via this generator. If None is given, the remover will not
               check for the number of already removed atoms before attempting
               a remove.
        """
        if maximumCollected is not None:
            assert is_integer(maximumCollected), LOGGER.error("maximumCollected must be an integer")
            maximumCollected = INT_TYPE(maximumCollected)
            assert maximumCollected>0, LOGGER.error("maximumCollected must be bigger than 0")
        self.__maximumCollected = maximumCollected

    def set_allow_fitting_scale_factor(self, allowFittingScaleFactor):
        """
        Set allow fitting scale factor flag.

        :Parameters:
            #. allowFittingScaleFactor (bool): Constraints and especially
               experimental ones have a scale factor constant that can be fit.
               Fitting a scale factor happens at stochastic engine's runtime
               at a certain fitting frequency. If this flag set to True,
               then fitting the scale factor will be allowed upon removing
               atoms. When set to False, fitting the constraint scale factor
               will be forbidden upon removing atoms. By default,
               allowFittingScaleFactor is set to False because it's more
               logical to allow removing only atoms that enhances the total
               standard error without rescaling the model's data.
        """
        assert isinstance(allowFittingScaleFactor, bool), LOGGER.error("allowFittingScaleFactor must be boolean")
        self.__allowFittingScaleFactor = allowFittingScaleFactor

    def set_atoms_list(self, atomsList):
        """
        Set atoms index list from which atoms will be picked to attempt removal.
        This method must be overloaded and not be called from this class but
        from its children. Otherwise a usage error will be raised.

        :Parameters:
            #. atomsList (None, list,set,tuple,np.ndarray): The list of atoms
               index to chose and remove from.
        """
        if atomsList is not None:
            C = _Container()
            # add container
            if not C.is_container('removeAtomsList'):
                C.add_container('removeAtomsList')
            # check if atomsList already defined
            loc = C.get_location_by_hint(atomsList)
            if loc is not None:
                atomsList = C.get_value(loc)
            else:
                assert isinstance(atomsList, (list,tuple,np.ndarray)), LOGGER.error("atomsList must be either a list or a numpy.array")
                CL = []
                for idx in atomsList:
                    assert is_integer(idx), LOGGER.error("atomsList items must be integers")
                    assert idx>=0, LOGGER.error("atomsList item must equal or bigger than 0")
                    CL.append(INT_TYPE(idx))
                setCL = set(CL)
                assert len(setCL) == len(CL), LOGGER.error("atomsList redundancy is not allowed")
                AL = np.array(CL, dtype=INT_TYPE)
                # add swapList to container
                C.set_value(container='removeAtomsList', value=AL, hint=atomsList)
                atomsList = AL
        # set atomsList attribute
        self.__atomsList = atomsList
        # reset collector state
        self._collectorState = None

    def move(self, coordinates):
        """
        Moves coordinates.
        This method must NOT be overloaded in MoveGenerator sub-classes.

        :Parameters:
            #. coordinates (np.ndarray): Not used here.
        """
        raise Exception(LOGGER.usage("%s '%s' is not allowed in removes generators"%(self.__class__.__name__,inspect.stack()[0][3])))

    def transform_coordinates(self, coordinates, argument):
        """
        This method must NOT be overloaded in MoveGenerator sub-classes.

        :Parameters:
            #. coordinates (np.ndarray): Not used here.
               the translation.
            #. argument (object): Not used here.
        """
        raise Exception(LOGGER.usage("%s '%s' is not allowed in removes generators"%(self.__class__.__name__,inspect.stack()[0][3])))

    def pick_from_list(self, engine):
        """
        This method must be overloaded in all RemoveGenerator sub-classes.

        :Parameters:
            #. engine (Engine): stochastic engine calling the method.
        """
        raise Exception(LOGGER.impl("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))


class SwapGenerator(MoveGenerator):
    """
    It is a particular move generator that instead of generating a
    move upon a group of atoms, it will exchange the group atom positions
    with other atoms from a defined swapList.
    Because the swapList can be big, swapGenerator can be assigned to
    multiple groups at the same time under the condition of all groups
    having the same length.\n

    During stochastic engine runtime, whenever a swap generator is encountered,
    all sophisticated selection recurrence modes such as (refining, exploring)
    will be reduced to simple recurrence.\n

    This class can't be instantiated but its sub-classes might be.

    :Parameters:
        #. group (None, Group): The group instance.
        #. swapLength (Integer): The swap length that defines the length of
           the group and the length of the every swap sub-list in swapList.
        #. swapList (None, List): List of atoms index.
           If None is given, no swapping or exchanging will be performed.
           If List is given, it must contain lists of atom indexes where every
           sub-list must have the same number of atoms as the group.
    """
    def __init__(self, group=None, swapLength=1, swapList=None):
        super(SwapGenerator, self).__init__(group=group)
        # set swap length
        self.set_swap_length(swapLength)
        # set swap list
        self.set_swap_list(swapList)
        #  initialize swapping variables
        self.__groupAtomsIndexes = None
        self.__swapAtomsIndexes  = None
        # reset collector state
        self._collectorState = None

    @property
    def swapLength(self):
        """ Swap length."""
        return self.__swapLength

    @property
    def swapList(self):
        """ Swap list."""
        return self.__swapList

    @property
    def groupAtomsIndexes (self):
        """ Last selected group atoms index."""
        return self.__groupAtomsIndexes

    @property
    def swapAtomsIndexes(self):
        """ Last swap atoms index."""
        return self.__swapAtomsIndexes

    def set_swap_length(self, swapLength):
        """
        Set swap length. it will empty and reset swaplist automatically.

        :Parameters:
            #. swapLength (Integer): The swap length that defines the length
               of the group and the length of the every swap sub-list in
               swapList.
        """
        assert is_integer(swapLength), LOGGER.error("swapLength must be an integer")
        swapLength = INT_TYPE(swapLength)
        assert swapLength>0, LOGGER.error("swapLength must be bigger than 0")
        self.__swapLength = swapLength
        self.__swapList   = ()
        # set uncollected atoms swapList
        self._remainingAtomsSwapList  = self.__swapList
        # reset collector state
        self._collectorState = None

    def set_group(self, group):
        """
        Set the MoveGenerator group.

        :Parameters:
            #. group (None, Group): group instance.
        """
        MoveGenerator.set_group(self, group)
        if self.group is not None:
            assert len(self.group) == self.__swapLength, LOGGER.error("SwapGenerator groups length must be equal to swapLength.")

    def set_swap_list(self, swapList):
        """
        Set the swap-list to exchange atoms position from.

        :Parameters:
            #. swapList (None, List): The list of atoms.\n
               If None is given, no swapping or exchanging will be performed.\n
               If List is given, it must contain lists of atom indexes where
               every sub-list length must be equal to swapLength.
        """
        C = _Container()
        # add container
        if not C.is_container('swapList'):
            C.add_container('swapList')
        # check if swapList already defined
        loc = C.get_location_by_hint(swapList)
        if loc is not None:
            self.__swapList = C.get_value(loc)
        elif swapList is None:
            self.__swapList = ()
        else:
            SL = []
            assert isinstance(swapList, (list,tuple)), LOGGER.error("swapList must be a list")
            for sl in swapList:
                assert isinstance(sl, (list,tuple)), LOGGER.error("swapList items must be a list")
                subSL = []
                for num in sl:
                    assert is_integer(num), LOGGER.error("swapList sub-list items must be integers")
                    num = INT_TYPE(num)
                    assert num>=0, LOGGER.error("swapList sub-list items must be positive")
                    subSL.append(num)
                assert len(set(subSL))==len(subSL), LOGGER.error("swapList items must not have any redundancy")
                if self.swapLength is not None:
                    assert len(subSL) == self.swapLength, LOGGER.error("swapList item length must be equal to swapLength")
                SL.append(np.array(subSL, dtype=INT_TYPE))
            self.__swapList = tuple(SL)
            # add swapList to container
            C.set_value(container='swapList', value=self.__swapList, hint=swapList)
        # set uncollected atoms swapList
        self._remainingAtomsSwapList  = self.__swapList
        # reset collector state
        self._collectorState = None

    def append_to_swap_list(self, subList):
        """
        Append a sub list to swap list.

        :Parameters:
            #. subList (List): The sub-list of atoms index to append
               to swapList.
        """
        assert isinstance(subList, (list,tuple)), LOGGER.error("subList must be a list")
        subSL = []
        for num in subList:
            assert is_integer(num), LOGGER.error("subList items must be integers")
            num = INT_TYPE(num)
            assert num>=0, LOGGER.error("subList items must be positive")
            subSL.append(num)
        assert len(set(subSL))==len(subSL), LOGGER.error("swapList items must not have any redundancy")
        assert len(subSL) == self.__swapLength, LOGGER.error("swapList item length must be equal to swapLength")
        # append
        self.__swapList = list(self.__swapList)
        subSL = np.array(subSL, dtype=INT_TYPE)
        self.__swapList.append(subSL)
        self.__swapList = tuple(self.__swapList)
        # set uncollected atoms swapList
        self._remainingAtomsSwapList  = self.__swapList
        # reset collector state
        self._collectorState = None

    def _set_remaining_atoms_swap_list(self, engine):
        collectorState = engine._atomsCollector.state
        # check engine's atomsCollector state
        if collectorState == self._collectorState or not len(engine._atomsCollector):
            self._collectorState = collectorState
            return
        C = _Container()
        # add container
        if not C.is_container('swapList'):
            C.add_container('swapList')
        # get swapList location
        loc = C.get_location_by_hint(self.swapList)
        # if location exists
        if loc is not None:
            remainingAtomsSwapList = C.get_value(loc)
            # it must be a dict
            if not isinstance(remainingAtomsSwapList, dict):
                remainingAtomsSwapList = None
            # collector state must be the same as engineCollectorState
            elif remainingAtomsSwapList['collectorState'] != collectorState:
                remainingAtomsSwapList = None
            # if same as engineCollectorState
            else:
                remainingAtomsSwapList = remainingAtomsSwapList['remainingAtomsSwapList']
        # if location doesn't exit
        else:
            remainingAtomsSwapList = None
        # in case swapList needs to be rebuilt
        if remainingAtomsSwapList is None:
            remainingAtomsSwapList = []
            for sl in self.swapList:
                if engine._atomsCollector.any_collected(sl):
                    continue
                remainingAtomsSwapList.append(sl)
            # add to container
            value = {'remainingAtomsSwapList':remainingAtomsSwapList, 'collectorState':collectorState}
            C.set_value(container='swapList', value=value, hint=self.swapList)
        # set remainingAtomsSwapList
        self._remainingAtomsSwapList = remainingAtomsSwapList
        # update collectorState
        self._collectorState = collectorState

    def get_ready_for_move(self, engine, groupAtomsIndexes):
        """
        Set the swap generator ready to perform a move. Unlike a normal
        move generator, swap generators will affect not only the selected
        atoms but other atoms as well. Therefore at stochastic engine runtime,
        selected atoms will be extended to all affected atoms by the swap.\n
        This method is called automatically upon stochastic engine runtime
        to ensure that all affect atoms with the swap are updated.

        :Parameters:
            #. engine (fullrmc.Engine): The stochastic engine calling for
               the move.
            #. groupAtomsIndexes (numpy.ndarray): The atoms index to swap.

        :Returns:
            #. indexes (numpy.ndarray): All the atoms involved in the swap move
               including the given groupAtomsIndexes.
        """
        # update and set _remainingAtomsSwapList and _collectorState
        self._set_remaining_atoms_swap_list(engine=engine)
        # select
        self.__groupAtomsIndexes = groupAtomsIndexes
        # check if existing atoms swap list is not empty. if not swap with itself.
        if len(self._remainingAtomsSwapList):
            self.__swapAtomsIndexes  = self._remainingAtomsSwapList[ randint(0,len(self._remainingAtomsSwapList)-1) ]
        else:
            self.__swapAtomsIndexes = self.__groupAtomsIndexes
        return np.concatenate( (self.__groupAtomsIndexes,self.__swapAtomsIndexes) )


class PathGenerator(MoveGenerator):
    """
    PathGenerator is a MoveGenerator sub-class where moves definitions
    are pre-stored in a path and get pulled out at every move step.\n

    This class can't be instantiated but its sub-classes might be.

    :Parameters:
        #. group (None, Group): The group instance.
        #. path (None, list): The list of moves.
        #. randomize (boolean): Whether to pull moves randomly from path or
           pull moves in order at every step.
    """

    def __init__(self, group=None, path=None, randomize=False):
        super(PathGenerator, self).__init__(group=group)
        # set path
        self.set_path(path)
        # set randomize
        self.set_randomize(randomize)
        # initialize flags
        self.__initialize_path_generator__()

    def __initialize_path_generator__(self):
        self.__step = 0

    @property
    def step(self):
        """ Current step number."""
        return self.__step

    @property
    def path(self):
        """ Path list of moves."""
        return self.__path

    @property
    def randomize(self):
        """ Randomize flag."""
        return self.__randomize

    def check_path(self, path):
        """
        Check the generator's path.\n

        This method must be overloaded in all PathGenerator sub-classes.

        :Parameters:
            #. path (list): The list of moves.
        """
        raise Exception(LOGGER.error("%s '%s' method must be overloaded"%(self.__class__.__name__,inspect.stack()[0][3])))

    def normalize_path(self, path):
        """
        Normalizes all path moves. It is called automatically upon
        set_path method is called.\n

        This method can be overloaded in all MoveGenerator sub-classes.

        :Parameters:
            #. path (list): The list of moves.

        :Returns:
            #. path (list): The list of moves.
        """
        return list(path)

    def set_path(self, path):
        """
        Set the moves path.

        :Parameters:
            #. path (list): The list of moves.
        """
        valid, message = self.check_path(path)
        if not valid:
            raise Exception(message)
        # normalize path
        self.__path = self.normalize_path( path )
        # reset generator
        self.__initialize_path_generator__()

    def set_randomize(self, randomize):
        """
        Set whether to randomize moves selection.

        :Parameters:
            #. randomize (boolean): Whether to pull moves randomly from path
               or pull moves in order at every step.
        """
        assert isinstance(randomize, bool), LOGGER.error("randomize must be boolean")
        self.__randomize = randomize

    def move(self, coordinates):
        """
        Move coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the transformation.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the transformation.
        """
        if self.__randomize:
            move = self.__path[ randint(0,len(self.__path)-1) ]
        else:
            move = self.__path[self.__step]
            self.__step = (self.__step+1)%len(self.__path)
        # perform the move
        return self.transform_coordinates(coordinates, argument=move)


class MoveGeneratorCombinator(MoveGenerator):
    """
    MoveGeneratorCombinator combines all moves of a list of MoveGenerators
    and applies it at once.

    :Parameters:
        #. group (None, Group): The constraint stochastic engine.
        #. combination (list): The list of MoveGenerator instances.
        #. shuffle (boolean): Whether to shuffle generator instances at
           every move or to combine moves in the list order.


    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Core.MoveGenerator import MoveGeneratorCombinator
        from fullrmc.Generators.Translations import TranslationGenerator
        from fullrmc.Generators.Rotations import RotationGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...

        ##### Define each group move generator as a combination of a translation and a rotation. #####
        # create recursive group selector. Recurrence is set to 20 with explore flag set to True.
        # shuffle is set to True which means that at every selection the order of move generation
        # is random. At one step a translation is performed prior to rotation and in another step
        # the rotation is performed at first.
        # selected from the collector.
        for g in ENGINE.groups:
            # create translation generator
            TMG = TranslationGenerator(amplitude=0.2)
            # create rotation generator only when group length is bigger than 1.
            if len(g)>1:
                RMG = RotationGenerator(amplitude=2)
                MG  = MoveGeneratorCombinator(collection=[TMG,RMG],shuffle=True)
            else:
                MG  = MoveGeneratorCombinator(collection=[TMG],shuffle=True)
            g.set_move_generator( MG )
    """

    def __init__(self, group=None, combination=None, shuffle=False):
        # set combination
        self.__combination = []
        # initialize
        super(MoveGeneratorCombinator, self).__init__(group=group)
        # set path
        self.set_combination(combination=combination)
        # set randomize
        self.set_shuffle(shuffle=shuffle)

    @property
    def shuffle(self):
        """ Shuffle flag."""
        return self.__shuffle

    @property
    def combination(self):
        """ Combination list of MoveGenerator instances."""
        return self.__combination

    def check_group(self, group):
        """
        Checks the generator's group.
        This methods always returns True because normally all combination
        MoveGenerator instances groups are checked.\n

        This method must NOT be overloaded unless needed.

        :Parameters:
            #. group (Group): the Group instance
        """
        return True, ""

    def set_group(self, group):
        """
        Set the MoveGenerator group.

        :Parameters:
            #. group (None, Group): group instance.
        """
        MoveGenerator.set_group(self, group)
        for mg in self.__combination:
            mg.set_group(group)

    def set_combination(self, combination):
        """
        Set the generators combination list.

        :Parameters:
            #. combination (list): The list of MoveGenerator instances.
        """
        assert isinstance(combination, (list,set,tuple)), LOGGER.error("combination must be a list")
        combination = list(combination)
        for c in combination:
            assert isinstance(c, MoveGenerator), LOGGER.error("every item in combination list must be a MoveGenerator instance")
            assert not isinstance(c, SwapGenerator), LOGGER.error("SwapGenerator is not allowed to be combined")
            assert not isinstance(c, SwapGenerator), LOGGER.error("RemoveGenerator is not allowed to be combined")
            c.set_group(self.group)
        self.__combination = combination

    def set_shuffle(self, shuffle):
        """
        Set whether to shuffle moves generator.

        :Parameters:
            #. shuffle (boolean): Whether to shuffle generator instances at
               every move or to combine moves in the list order.
        """
        assert isinstance(shuffle, bool), LOGGER.error("shuffle must be boolean")
        self.__shuffle = shuffle

    def move(self, coordinates):
        """
        Move coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the transformation.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the transformation.
        """
        indexes = range(len(self.__combination))
        if self.__shuffle:
            shuffle( indexes )
        # create the move combination
        for idx in indexes:
            coordinates = self.__combination[idx].move(coordinates)
        return coordinates


class MoveGeneratorCollector(MoveGenerator):
    """
    MoveGeneratorCollector collects MoveGenerators instances and applies
    the move of one instance at every step.

    :Parameters:
        #. group (None, Group): The constraint stochastic engine.
        #. collection (list): The list of MoveGenerator instances.
        #. randomize (boolean): Whether to pull MoveGenerator instance
           randomly from collection list or in order.
        #. weights (None, list): Generators selections Weights list.
           It must be None for equivalent weighting or list of
           (generatorIndex, weight) tuples.
           If randomize is False, weights list is ignored upon generator
           selection from collection.

    .. code-block:: python

        # import fullrmc modules
        from fullrmc.Engine import Engine
        from fullrmc.Core.MoveGenerator import MoveGeneratorCollector
        from fullrmc.Generators.Translations import TranslationGenerator
        from fullrmc.Generators.Rotations import RotationGenerator

        # create engine
        ENGINE = Engine(path='my_engine.rmc')

        # set pdb file
        ENGINE.set_pdb('system.pdb')

        # Add constraints ...
        # Re-define groups if needed ...

        ##### Define each group move generator as a combination of a translation and a rotation. #####
        # create recursive group selector. Recurrence is set to 20 with explore flag set to True.
        # randomize is set to True which means that at every selection a generator is randomly
        # selected from the collector.
        for g in ENGINE.groups:
            # create translation generator
            TMG = TranslationGenerator(amplitude=0.2)
            # create rotation generator only when group length is bigger than 1.
            if len(g)>1:
                RMG = RotationGenerator(amplitude=2)
                MG  = MoveGeneratorCollector(collection=[TMG,RMG],randomize=True)
            else:
                MG  = MoveGeneratorCollector(collection=[TMG],randomize=True)
            g.set_move_generator( MG )

    """
    def __init__(self, group=None, collection=None, randomize=True, weights=None):
        # set collection
        self.__collection = []
        # initialize
        super(MoveGeneratorCollector, self).__init__(group=group)
        # set path
        self.set_collection(collection)
        # set randomize
        self.set_randomize(randomize)
        # set weights
        self.set_weights(weights)
        # initialize flags
        self.__initialize_generator()

    def __initialize_generator(self):
        self.__step = 0

    def __check_single_weight(self, w):
        """Check a single group weight tuple format."""
        assert isinstance(w, (list,set,tuple)),LOGGER.error("weights list items must be tuples")
        assert len(w)==2, LOGGER.error("weights list tuples must have exactly 2 items")
        idx  = w[0]
        wgt = w[1]
        assert is_integer(idx), LOGGER.error("weights list tuples first item must be an integer")
        idx = INT_TYPE(idx)
        assert idx>=0, LOGGER.error("weights list tuples first item must be positive")
        assert idx<len(self.__collection), LOGGER.error("weights list tuples first item must be smaller than the number of generators in collection")
        assert is_number(wgt), LOGGER.error("weights list tuples second item must be an integer")
        wgt = FLOAT_TYPE(wgt)
        assert wgt>0, LOGGER.error("weights list tuples first item must be bigger than 0")
        # all True return idx and weight
        return idx, wgt

    @property
    def randomize(self):
        """ Randomize flag."""
        return self.__randomize

    @property
    def collection(self):
        """ List of MoveGenerator instances."""
        return self.__collection

    @property
    def generatorsWeight(self):
        """ Generators selection weights list."""
        return self.__generatorsWeight

    @property
    def selectionScheme(self):
        """ Selection scheme."""
        return self.__selectionScheme

    def set_group(self, group):
        """
        Set the MoveGenerator group.

        :Parameters:
            #. group (None, Group): group instance.
        """
        MoveGenerator.set_group(self, group)
        for mg in self.__collection:
            mg.set_group(group)

    def check_group(self, group):
        """
        Check the generator's group. This methods always returns True
        because normally all collection MoveGenerator instances groups
        are checked.\n

        This method must NOT be overloaded unless needed.

        :Parameters:
            #. group (Group): the Group instance.
        """
        return True, ""

    def set_collection(self, collection):
        """
        Set the generators instances collection list.

        :Parameters:
            #. collection (list): The list of move generator instance.
        """
        assert isinstance(collection, (list,set,tuple)), LOGGER.error("collection must be a list")
        collection = list(collection)
        for c in collection:
            assert isinstance(c, MoveGenerator), LOGGER.error("every item in collection list must be a MoveGenerator instance")
            assert not isinstance(c, SwapGenerator), LOGGER.error("SwapGenerator is not allowed to be collected")
            assert not isinstance(c, SwapGenerator), LOGGER.error("RemoveGenerator is not allowed to be collected")
            c.set_group(self.group)
        self.__collection = collection
        # reset generator
        self.__initialize_generator()

    def set_randomize(self, randomize):
        """
        Set whether to randomize MoveGenerator instance selection
        from collection list.

        :Parameters:
            #. randomize (boolean): Whether to pull MoveGenerator instance
               randomly from collection list or in order.
        """
        assert isinstance(randomize, bool), LOGGER.error("randomize must be boolean")
        self.__randomize = randomize

    def set_weights(self, weights):
        """
        Set groups selection weighting scheme.

        :Parameters:
            #. weights (None, list): Generators selections Weights list.
               It must be None for equivalent weighting or list of
               (generatorIndex, weight) tuples.
               If randomize is False, weights list is ignored upon generator
               selection from collection.
        """
        generatorsWeight = np.ones(len(self.__collection), dtype=FLOAT_TYPE)
        if weights is not None:
            assert isinstance(weights, (list,set,tuple)),LOGGER.error("weights must be a list")
            for w in weights:
                idx, wgt = self.__check_single_weight(w)
                # update groups weight
                generatorsWeight[idx] = wgt
        # set groups weight
        self.__generatorsWeight = generatorsWeight
        # create selection histogram
        self.set_selection_scheme()

    def set_selection_scheme(self):
        """ Set selection scheme. """
        cumsumWeights = np.cumsum(self.__generatorsWeight, dtype=FLOAT_TYPE)
        self.__selectionScheme = cumsumWeights/cumsumWeights[-1]

    def move(self, coordinates):
        """
        Move coordinates.

        :Parameters:
            #. coordinates (np.ndarray): The coordinates on which to apply
               the transformation.

        :Returns:
            #. coordinates (np.ndarray): The new coordinates after applying
               the transformation.
        """
        if self.__randomize:
            index = INT_TYPE( np.searchsorted(self.__selectionScheme, generate_random_float()) )
            moveGenerator = self.__collection[ index ]
        else:
            moveGenerator = self.__collection[self.__step]
            self.__step   = (self.__step+1)%len(self.__collection)
        # perform the move
        return moveGenerator.move(coordinates)
